#!/usr/bin/env python
# -*- coding: utf8 -*-

import sys
import os
import re
import json
from collections import defaultdict
from datetime import datetime
import numpy

from fmqlutils.cacher.cacherUtils import SCHEMA_LOCN_TEMPL 
from fmqlutils.reporter.reportUtils import MarkdownTable, reportPercent, reportAbsAndPercent

from ..webReportUtils import SITE_DIR_TEMPL, TOP_MD_TEMPL

"""
Institutions: https://www.va.gov/directory/Guide/fac_list_by_state.cfm?State=OH&dnum=All

TODO: some lookup of name + DESCRIBE TYPE for 3.081 supplies last day
"""

TEMPL = """## Introduction

Today the Veterans Administration (VA) provides and manages veteran care using <span style='yellowIt'>130</span> deployments of their in-house VistA system. Each VistA is an integrated suite of over one hundred applications sharing a single hierarchal database, FileMan. In deployed VistAs, FileMan holds up to <span style='yellowIt'>2 billion</span> interlinked records of over <span style='yellowIt'>4000</span> different types.

As well as the vast majority of Veteran clinical data, FileMan holds profiles of VA users, financial information, connected device descriptions, interface definitions and usage logs. The integrated and comprehensive nature of FileMan data makes it an ideal target for analysis to inform the migration of veteran care from VistA and into EHRM.

This __\"VistA by the Numbers\"__ report is one of a series, each of which captures the activity of a specific active VA VistA system. Each is generated by a framework for VistA analytics that first caches relevant data from a FileMan and then generates appropriate system descriptions. Such automation ensures common criteria for analysis and the ability to cover all <span style='yellowIt'>130</span> systems deployed in VA even as additional criteria are added to the framework over time.

This report describes __{}__ which has VA identifier or station number {}. Specifically, it describes a clone cut on {} and housed in the VAEC.
"""

def webReportIntro(stationNo):

    systemName = "{NEED NAME}" # ex/ Mann Grandstaff (Spokane) VistA
    
    lastSignInDay = "" # NEED DESCRIBE TYPE of 3.081 with "lastIEN"

    mu = TEMPL.format(
        systemName,
        stationNo,
        lastSignInDay
    )

# ################################# DRIVER #######################
               
def main():

    assert(sys.version_info >= (2,7))

    if len(sys.argv) < 2:
        print "need to specify station # ex/ 442 - exiting"
        return
        
    stationNo = sys.argv[1]
    
    userSiteDir = SITE_DIR_TEMPL.format(stationNo)
    if not os.path.isdir(userSiteDir):
        raise Exception("Expect User Site to already exist with its basic contents")

    webReportIntro(stationNo)
    
if __name__ == "__main__":
    main()


#!/usr/bin/env python
# -*- coding: utf8 -*-

import sys
import os
import re
import json
from collections import defaultdict, Counter

import numpy
import math

from fmqlutils.cacher.cacherUtils import FMQLReplyStore, FilteredResultIterator, SCHEMA_LOCN_TEMPL, DATA_LOCN_TEMPL, DATARF_LOCN_TEMPL, TMPWORKING_LOCN_TEMPL
from fmqlutils.typer.reduceReportTypes import TYPER_REDUCTIONS_LOCN_TEMPL
from fmqlutils.reporter.reportUtils import MarkdownTable

"""
Site
"""
SITE_DIR_TEMPL = "/data/vista/{}/ReportSite/" 

TOP_MD_TEMPL = """---
layout: default
title: {}
---

"""

def ensureWebReportLocations(stationNo):
    reportSiteLocation = SITE_DIR_TEMPL.format(stationNo)
    if not os.path.isdir(reportSiteLocation):
        os.mkdir(reportSiteLocation)

def dtdd(dt, dd):
    return "{}\n:    {}\n\n".format(dt, dd)
    
"""
Hide need to include DATA_LOCN_TEMPL and SCHEMA
"""
def filteredResultIterator(stationNo, flTyp, useRF=False, startAtReply=""):
    locn = DATARF_LOCN_TEMPL.format(stationNo) if useRF else DATA_LOCN_TEMPL.format(stationNo)
    return FilteredResultIterator(locn, flTyp, startAtReply=startAtReply)
    
def loadSelectTypes(stationNo):
    return  json.load(open(SCHEMA_LOCN_TEMPL.format(stationNo) + "SELECT_TYPES.json"))

# ############################### Stat Utils #########################

"""
Shows off key stats with a little (obvious) stats observations

https://docs.scipy.org/doc/numpy/user/

TODO: more on numpy (ie/ go through it as go through books)
"""
def reportKeyStats(kstats):
    print("Count Values: {}".format(kstats["count"]))
    mn = kstats["mean"]
    print("Mean: {}".format(mn))
    M = kstats["median"]
    print("Median (M): {}".format(M))
    if M > mn:
        print("\t... M bigger than Mean so data skews left")
    elif M < mn:
        print("\t... M smaller than Mean so data skews right")
    print("Standard Deviation (s): {}".format(kstats["std"]))
    print("\treflect skew of min {} and max {}".format(kstats["min"], kstats["max"]))
    print("IQR: {}".format(kstats["iqr"]))                                                   
    print("\tbetween 1st ({}) and 3rd ({}) quartiles".format(kstats["quartileOne"], kstats["quartileThree"]))
    if "olt" in kstats:
        print("Outliers below", kstats["olt"])
    print("Outliers above\n", kstats["oht"])
    
"""
Key stats for distribution of values

min, max
Quartile: 1, 2 (Median), 3 for data-spread
... "five-number summary" ideal for data w/ outliers and a 'box-and-whisker' plot 
... Quartiles divide a rank-ordered data set into four equal parts at 25, 50, 75
percentiles
IQR
... "midspread" - best at spread if outliers
olt < (Q1-1.5*IQR)
oht > Q3+1.5*IQR
... for key outliers: < are lows, > are highs. Only present if > max, < min
mean
std
... for symm distributions with no outliers where five numbers don't kick in

TODO: consider <span class='yellowIt'>{:.2f}%</span> and numpy.around or
roundFloat (see below)

oht: 1.5 == "inner fences"
major outliers: Q3 + 3*IQR == "outer fences" ... need name for this.

Note: this is 'Univariate method' (See: https://www.kdnuggets.com/2017/01/3-methods-deal-outliers.html / 'Minkowski error' etc)

https://en.wikipedia.org/wiki/Outlier (Tukey's fences)
- above: Q3 + K(Q3-Q1) where K is 1.5 and K = 3 is "far out"
- below: Q1 - K(Q1-Q3)
ie/ K times Quartile. Fences for box plots.
"""
def keyStats(values):
    quartileOne = numpy.percentile(values, 25)
    median = numpy.percentile(values, 50)
    quartileThree = numpy.percentile(values, 75)
    # get variance in middle 50% of data
    # numpy.subtract(*numpy.percentile(values, [75, 25]))
    iqr = quartileThree - quartileOne # interquartile range    
    oht = quartileThree + (1.5 * iqr) # Outlier High above this
    cntValsGTOHT = sum(1 for val in values if val > oht) 
    ohto = quartileThree + (3 * iqr) # idea of "outer fence": TODO: get name
    cntValsGTOHTO = sum(1 for val in values if val > ohto) 
    stats = {"total": sum(val for val in values), "count": len(values), "min": numpy.min(values), "quartileOne": quartileOne, "median": median, "quartileThree": quartileThree, "max": numpy.max(values), "iqr": iqr, "oht": oht, "ohto": ohto, "mean": numpy.mean(values), "std": numpy.std(values)}
    olt = quartileOne - (1.5 * iqr) # Outlier Low below this
    if int(olt) > 0: # only olt if > 0 ie/ negative ie/ all acceptable
        stats["olt"] = olt
    """
    Own: should show outliers? only go into outliers if big ie/ if pretty smooth 
    """
    stats["hasOutliers"] = True if stats["max"] - stats["oht"] > stats["median"] else False
    if cntValsGTOHT:
        stats["gtoht"] = cntValsGTOHT
    if cntValsGTOHTO:
        stats["gtohto"] = cntValsGTOHTO
    return stats
    
"""
Use typically for byValueCount off typer or a Counter

    "a": 1
    "b": 2
    "c": 1
    
-----> [a, b, b, c]

and can pass into kstats if forceInt

Note: forceInt is for the measured values which may be integers but
are in string form.
"""
def flattenFrequencyDistribution(freqDistrib, forceInt=True):
    freqDistribFlat = []
    for entry in freqDistrib:
        f = [int(entry) if forceInt else entry for i in range(int(freqDistrib[entry]))]
        freqDistribFlat.extend(f)
    return freqDistribFlat
    
"""
New Stats to expand:

- (basic) and relative (percent/ration) frequency distribution ie/ cnt occurrences
        from collections import Counter
        Counter([...]) ... 
  ... put relatives into a stacked bar chart
  ... conditional => a number of freq distribs, one each per condition (ex/ for PDFs, WSs)
- bar (or a pie) to histogram: when # of distict values too big then go to histogram
- null hypothesis/expected frequencies: variables unrelated (type of image and WS?)
  - https://www.spss-tutorials.com/chi-square-independence-test/
"""

"""
- cumulative frequencies add as go (https://www.spss-tutorials.com/frequency-distribution-what-is-it/) ... no good for nominals ("French" or "PDF") but for good/better/best ...
  ie age of patients  ... https://www.wyzant.com/resources/lessons/math/statistics_and_probability/averages/cumulative_frequency_percentiles_and_quartiles
  ... A cumulative frequency graph, also known as an Ogive
  ... good site for stuff.
  ... cumulative freq can be used to gen quartiles
"""        

"""
Clustering, numpy has it (ntlk uses it - but mainly see intelligent web book)
... k ... and clumping 'users' or ... based on similarity/distance.
"""

"""
after excel, try matplotlib, networkx 
"""
    
# ############################ Simple Report Utils #################

"""
If X.0 -> 0 otherwise to 2 pts (or whatever is passed in)
"""
def roundFloat(no, to=2):
    no = no.round(to)
    if no.is_integer():
        return int(no)
    return no
    
# ############################ Plot Utils ##########################

"""
Crude Example of Graph

import matplotlib.pyplot as plt

See: https://python-graph-gallery.com/120-line-chart-with-matplotlib/ and
https://matplotlib.org/gallery/index.html

    plot([2007, 2008, 2009, 2010], [1, 2, 3, 4])

"""
def plot(xs, ys):

    import matplotlib.pyplot as plt
    plt.plot(xs, ys, "ro")
    plt.xlabel("Time")
    plt.ylabel("Ibs")
    plt.show()
            
# ######################### Reduce Users ###########################

"""
Returns Users, Indexed 

Note: RF part ie/ 200 into DataRF will move out of here. Tmp index
maker for any type (including RF types) will become one simple utility.

TODO: go to lookup by IEN as simpler
"""
def reduce200(stationNo, onlyUsers=None, forceNew=False, coreOnly=False):

    jsnFl = TMPWORKING_LOCN_TEMPL.format(stationNo) + "userReduction{}.json".format("CORE" if coreOnly else "")
    
    if not forceNew:
        try:
            infosByRef = json.load(open(jsnFl))
        except:
            pass
        else:
            return infosByRef

    print("First time through, creating index by id for Users (200) for {}".format(stationNo))
    
    dataLocn = DATA_LOCN_TEMPL.format(stationNo)
    resourceIter = FilteredResultIterator(dataLocn, "200")
    infosByRef = {}
    cnt = 0
    for resource in resourceIter:
        cnt += 1
                
        if cnt % 1000 == 0:
            print("Processed another 1000 users - now at {:,}".format(len(infosByRef)))
        
        if onlyUsers and resource["_id"] not in onlyUsers:
            continue
        
        userId = "{} [{}]".format(resource["label"], resource["_id"])
                
        info = {"hasProps": sorted([key for key in resource.keys() if key not in set(["_id", "label", "type"])])}
        
        if "creator" in resource:
            if resource["creator"]["id"] == "200-0":
                info["is_created_by_0"] = True
            else:
                createdId = "{} [{}]".format(resource["creator"]["label"], resource["creator"]["id"])
                info["creator"] = createdId
        
        if "date_entered" in resource:
            info["date_entered"] = resource["date_entered"]["value"]
            
        if "title" in resource:
            info["title"] = resource["title"]["label"] # it's a pointer
            
        if "ssn" in resource:
            info["ssn"] = resource["ssn"]
            
        # used for lookup BEFORE SSN in setup. goes with subject_organizations etc
        if "secid" in resource:
            info["secid"] = resource["secid"]
            if "subject_organization" in resource and resource["subject_organization"] != "Department Of Veterans Affairs":
                raise Exception("Expected Subject Org for sec id to always be VA!")
            
        if "termination_date" in resource:
            info["termination_date"] = resource["termination_date"]["value"]
        if "termination_reason" in resource:
            info["termination_reason"] = resource["termination_reason"] 
            
        if "network_username" in resource:
            info["network_username"] = resource["network_username"]   
            
        if "last_signon_date_time" in resource:
            info["last_signon_date_time"] = resource["last_signon_date_time"]["value"]
            
        infosByRef[userId] = info 
            
        if coreOnly:
            continue
                            
        if "user_class" in resource:
            if len(resource["user_class"]) > 1:
                raise Exception("Expected only one user class assertion - ie/ singleton multiple")
            info["user_class"] = [uci["user_class"]["label"] for uci in resource["user_class"]][0]
            
        if "file_manager_access_code" in resource and resource["file_manager_access_code"] == "@":
            info["has_fileman_programmer_access"] = True
                
        if "primary_menu_option" in resource:
            info["primary_menu_option"] = resource["primary_menu_option"]["label"]
                
        if "secondary_menu_options" in resource:
            smos = set() # allows for dups with crude SMO addition
            for smo in resource["secondary_menu_options"]:
                if not re.match(r'\d+$', smo["secondary_menu_options"]["label"]):
                    smos.add(smo["secondary_menu_options"]["label"])
            if len(smos):
                info["secondary_menu_options"] = sorted(list(smos))
            
        if "delegated_options" in resource:
            dos = set()
            for delo in resource["delegated_options"]:
                if not re.match(r'\d+$', delo["delegated_options"]["label"]):
                    dos.add(delo["delegated_options"]["label"])
            if len(dos):                   
                info["delegated_options"] = sorted(list(dos))
            
        if "keys" in resource:
            keys = set() 
            for key in resource["keys"]:
                if not re.match(r'\d+$', key["key"]["label"]):
                    keys.add(key["key"]["label"])
            if len(keys):
                info["keys"] = sorted(list(keys))   
                
        if "preferred_editor" in resource: # for REMOTE user checks
            info["preferred_editor"] = resource["preferred_editor"]["label"]
            
        if "terminal_type_last_used" in resource:
            info["terminal_type_last_used"] = resource["terminal_type_last_used"]["label"] 
            
        # Consider "first sign on"
        if "visited_from" in resource:
            finalVisitedTime = None
            visitedFroms = {}
            for i, mresource in enumerate(resource["visited_from"]):
                if "visited_from" in mresource: 
                    visitedFrom = re.sub(r'[^\d]+', '', mresource["visited_from"])
                    duzAtHomeSite = mresource["duz_at_home_site"] if "duz_at_home_site" in mresource else ""
                    siteName = mresource["site_name"] if "site_name" in mresource else ""
                    siteId = ":".join([visitedFrom, duzAtHomeSite, siteName])
                    details = None
                    if "first_visit" in mresource or "last_visited" in mresource:
                        details = {}
                        if "first_visit" in mresource:
                            details["first"] = mresource["first_visit"]["value"]
                        if "last_visited" in mresource:
                            details["last"] = mresource["last_visited"]["value"]                            
                    visitedFroms[siteId] = details
            if len(visitedFroms):
                info["signed_on_froms"] = visitedFroms
            
    print("Serializing and Returning Index of {:,} Users (200)".format(cnt))
    if coreOnly:
        print("\t... core only")

    json.dump(infosByRef, open(jsnFl, "w"), indent=4)
    
    return infosByRef
    
# ############################ Reduce Patient ########################

def reduce2(stationNo, forceNew=False):

    jsnFl = TMPWORKING_LOCN_TEMPL.format(stationNo) + "patientReduction.json"
    
    if not forceNew:
        try:
            infosByIEN = json.load(open(jsnFl))
        except:
            pass
        else:
            return infosByIEN

    print("First time through, creating patient entered reduction for {}".format(stationNo))
    sys.stdout.flush()
    
    dataLocn = DATA_LOCN_TEMPL.format(stationNo)
    resourceIter = FilteredResultIterator(dataLocn, "2")
    infosByIEN = {}
    PROPS_TO_TAKE = set([
        "name", 
        "sex", 
        "date_of_birth", 
        "social_security_number", 
        
        "who_entered_patient", 
        "date_entered_into_file", 
        
        "city", 
        "state", 
        "zip_code", 
        
        "period_of_service", 
        
        "date_of_death", 
        "death_entered_by", 
        
        "primary_eligibility_code", 
        "combat_service_indicated", 
        
        "preferred_facility", 
        
        # TODO: expand these - ensure all id's used by VA/C are in here
        "integration_control_number", 
        "subscription_control_number", # 774 entries for CIRN (see if sync file
        "full_icn"    
    ])
    for resource in resourceIter:
        patientIEN = resource["_id"].split("-")[1]
        info = {"ien": patientIEN}
        for prop in PROPS_TO_TAKE:
            if prop in resource:
                if isinstance(resource[prop], list):
                    raise Exception("Not reducing multiples of 2")
                if isinstance(resource[prop], dict):
                    if "value" in resource[prop]:
                        info[prop] = resource[prop]["value"]
                    else:
                        info[prop] = resource[prop]["label"]
                else: 
                    info[prop] = resource[prop]
        infosByIEN[patientIEN] = info
        if len(infosByIEN) % 1000 == 0:
            print("Processed another 1000 patients - now at {:,}".format(len(infosByIEN)))
            sys.stdout.flush()
    
    print("Serializing and Returning Index of {:,} Users (2)".format(len(infosByIEN)))
    sys.stdout.flush()

    json.dump(infosByIEN, open(jsnFl, "w"), indent=4)
    
    return infosByIEN
    
# ############################ Reduce Location (44) ###########################

"""
Like other "anchor file" reductions, preludes to more formal reframes (RF) that
split out different types of key data into distinct types AND establishes a
lookup 

<----- will see if AppointmentHospitalLocation can be a sub class or split out appt info

Note: allows RF take - right now that RF is just the result of extracting the
44.001 and other multiples.
"""
def reduce44(stationNo, useRF=False, forceNew=False): 

    """
                for litProp in ["patient_friendly_name", "physical_location", "length_of_appointment", "default_appointment_type"]: # includes enum
                if litProp not in resource:
                    continue
                red[litProp] = resource[litProp]
            for dtProp in ["inactivate_date", "reactivate_date"]:
                if dtProp not in resource:
                    continue
                red[dtProp] = resource[dtProp]["value"]
            for refProp in ["stop_code_number", "institution", "division", "type_extension", "default_provider"]:
                if refProp not in resource:
                    continue
                    
    Add mult of provider (list) etc and the letter props
    """
    pass

    # TODO: out of custom reports
    
# ############################ Simple Reusable MD Makers ###############################

def mdFilesOfDomain(stationNo, domainName, bottomFileNo, beyondFileNo, usedFiles):

    selectTypes = loadSelectTypes(stationNo)

    fileTypeInfos = [typeInfo for typeInfo in selectTypes["results"] if "number" in typeInfo and float(typeInfo["number"]) >= bottomFileNo and float(typeInfo["number"]) < beyondFileNo and "count" in typeInfo]
     
    mu = """There are <span class="yellowIt">{:,}</span> files in the system dedicated to the _{}_ ...
    
""".format(len(fileTypeInfos), domainName)
    
    tbl = MarkdownTable(["Name", "\#", "Count"])
    for typeInfo in fileTypeInfos:
        tbl.addRow([
                "__{}__".format(typeInfo["name"]) if typeInfo["name"] in usedFiles else typeInfo["name"], 
            typeInfo["number"], 
            typeInfo["count"]
    ])
    mu += tbl.md() + "\n\n"
        
    return mu
    
# ######################### Simple Type Utils ########################

class PropertyCounter:
    
    def __init__(self):
        self.__counts = Counter()
        self.__countResources = 0
            
    def resourceProperties(self, props):
        for prop in props:
            self.__counts[prop] += 1
        self.__countResources += 1
            
    def mandatories(self):
        return sorted([prop for prop in self.__counts if self.__counts[prop] == self.__countResources]) 
            
    def optionalCounts(self):
        return dict((prop, self.__counts[prop]) for prop in self.__counts if self.__counts[prop] < self.__countResources) 



#!/usr/bin/env python
# -*- coding: utf8 -*-

import sys
import os
import re
import json
from collections import defaultdict
from datetime import datetime

from fmqlutils.reporter.reportUtils import MarkdownTable, reportPercent, reportAbsAndPercent

from fmqlutils.typer.reduceTypeUtils import splitTypeDatas, checkDataPresent, singleValue

from fmqlreports.webReportUtils import SITE_DIR_TEMPL

"""
Basic Patient Enrollment

Requires:
- 27_11 ?

... needs to be fully filled in
"""



#!/usr/bin/env python
# -*- coding: utf8 -*-

import sys
import os
import re
import json
from collections import defaultdict, OrderedDict, Counter
from datetime import datetime, date
import numpy

from fmqlutils.reporter.reportUtils import MarkdownTable, reportPercent, reportAbsAndPercent

from fmqlutils.typer.reduceTypeUtils import splitTypeDatas, checkDataPresent, singleValue, combineSubTypes, muBVC, muBVCOfSTProp
from fmqlutils.cacher.cacherUtils import metaOfVistA
from ..webReportUtils import ensureWebReportLocations, TOP_MD_TEMPL, SITE_DIR_TEMPL, keyStats, flattenFrequencyDistribution, roundFloat, reduce200
from userClassifier import UserClassifier

"""
Basic User Type Summary/Overview

Unlike the stopcode centric reports (off locations), these are used and institution centered. Approach is PER USER STs and then Categorize Users (=> STs) as REMOTE
etc.

Requires:
- 3.081 (Type Reduction by User: YR1 unless 'smallSOSet')
- 200 (All/Indexed-reduce200)
- ... Will move to require image logs too to round out

TODO: classification is work in progress - remotes from non key institution, locals
from non subordinate divisions etc.

Note: smallSOSet setup for case where only have small sample of SOs (ie/ take all)  
"""
def webReportUser(stationNo, deidentify=False, smallSOSet=False):

    print("Preparing to make user report for {}{} - loading data ...".format(stationNo, " [DEIDENTIFY]" if deidentify else ""))

    allThere, details = checkDataPresent(stationNo, [
        # Additional settings: force count on 
        #     ipv4_address, remote_station_id, remote_user_ien
        {"fileType": "3_081", "check": "YR1" if not smallSOSet else "TYPE"},
        {"fileType": "200", "check": "ALL"} # Index
    ])
    if not allThere:
        raise Exception("Some required data is missing - {}".format(details))
        
    ensureWebReportLocations(stationNo)
        
    meta = metaOfVistA(stationNo)
    vistaName = "VistA" if "name" not in meta else meta["name"]

    mu = TOP_MD_TEMPL.format("{} Users".format(stationNo), "Users")

    userInfoByUserRef = reduce200(stationNo)    
    type3_081, st3_081s = splitTypeDatas(stationNo, "3_081", reductionLabel="" if smallSOSet else "YR1", expectSubTypeProperty="user")
    st3_081ByUserRef = dict((singleValue(st, "user"), st) for st in st3_081s if "user" in st)
    
    userClassifier = UserClassifier(stationNo, userInfoByUserRef, type3_081, st3_081ByUserRef)
    print("Classifying Users ...")
    classification = userClassifier.classify()
    print("... Classification complete")

    if "_createDateProp" not in type3_081:
        raise Exception("YR1 3.081's must have create date prop")
    createDatePropInfo = type3_081[type3_081["_createDateProp"]]
    overallFirstSODate = createDatePropInfo["firstCreateDate"].split("T")[0]
    overallLastSODate = createDatePropInfo["lastCreateDate"].split("T")[0]
        
    BLURB_TEMPL = """Of the <span class='yellowIt'>{:,}</span> users known to {}, a minority <span class='yellowIt'>{}</span> ("__Active Users__") signed in between {} and {}, {}.

Most Active Users are real people but there are "__Machine Users__" for specialized VA batch applications or applications which manage end users themselves.

"""

    mu += BLURB_TEMPL.format(
        len(userInfoByUserRef),
        "{} [{}]".format(vistaName, stationNo),
        reportAbsAndPercent(len(classification["activeUserRefs"]), len(userInfoByUserRef)),
        overallFirstSODate, 
        overallLastSODate,
        "the period for which sign on logs are available" if smallSOSet else "the most recent year for which data is available"
    )  
    
    if deidentify:
        mu += "__DE-IDENTIFIED__: the names and national identifiers of real (Non-Machine) end users have been scrubbed from this report. VistA-specific IENs are left to identify such users for those with access to the system.\n\n"
            
    tbl = MarkdownTable([":Type", "Users", "SignOns"], includeNo=False)
    tbl.addRow([
        "Total Users (200 entries)", 
        len(userInfoByUserRef),
        ""
    ])
    tbl.addRow([
        "Active Users (3.081s for period)", 
        reportAbsAndPercent(len(classification["activeUserRefs"]), len(userInfoByUserRef)),
        type3_081["_total"]
    ])
    tbl.addRow([
        "DoD User",
        muUserRef(classification["dodUserRef"]),
        reportAbsAndPercent(
            st3_081ByUserRef[classification["dodUserRef"]]["_total"],
            type3_081["_total"]
        )
    ])
    tbl.addRow([
        "Postmaster",
        muUserRef(classification["postmasterUserRef"]),
        reportAbsAndPercent(
            st3_081ByUserRef[classification["postmasterUserRef"]]["_total"],
            type3_081["_total"]
        )
    ])
    tbl.addRow([
        "Active __Proxy Users__", 
        len(classification["activeProxyUserRefs"]),
        reportAbsAndPercent(
            sum(st3_081ByUserRef[userRef]["_total"] for userRef
 in classification["activeProxyUserRefs"]),
            type3_081["_total"]
        )
    ])  
    tbl.addRow([
        "Active __(Non Proxy) Machine Users__", 
        len(classification["activeNonProxyMachineUserRefs"]),
        reportAbsAndPercent(
            sum(st3_081ByUserRef[userRef]["_total"] for userRef
 in classification["activeNonProxyMachineUserRefs"]),
            type3_081["_total"]
        )
    ])
    tbl.addRow([
        "Active __Remote Users__", 
        reportAbsAndPercent(len(classification["activeRemoteUserRefs"]), len(classification["activeUserRefs"])),
        reportAbsAndPercent(
            sum(st3_081ByUserRef[userRef]["_total"] for userRef
 in classification["activeRemoteUserRefs"]),
            type3_081["_total"]
        )
    ])
    tbl.addRow([
        "Active __Local Users__", 
        reportAbsAndPercent(len(classification["activeLocalUserRefs"]), len(classification["activeUserRefs"])),
        reportAbsAndPercent(
            sum(st3_081ByUserRef[userRef]["_total"] for userRef
 in classification["activeLocalUserRefs"]),
            type3_081["_total"]
        )
    ])
    mu += "User type signon summary ...\n\n"
    mu += tbl.md() + "\n\n"
    
    mu += "Signon by week day ...\n\n"
    mu += tblByWeekDay(type3_081) + "\n\n"
    
    mu += webReportPostmaster(
        classification["postmasterUserRef"],
        userInfoByUserRef[classification["postmasterUserRef"]],
        type3_081["_total"],
        st3_081ByUserRef[classification["postmasterUserRef"]],
        classification["warningsByUserRef"][classification["postmasterUserRef"]] if classification["postmasterUserRef"] in classification["warningsByUserRef"] else []
    )
    
    mu += webReportDoDUser(
        classification["dodUserRef"],
        userInfoByUserRef[classification["dodUserRef"]],
        type3_081["_total"],
        st3_081ByUserRef[classification["dodUserRef"]],
        classification["warningsByUserRef"][classification["dodUserRef"]] if classification["dodUserRef"] in classification["warningsByUserRef"] else [],
        deidentify
    )
    
    # Beside's the special 2, 4 classes of User
    
    mu += webReportProxyUsers(
        classification["activeProxyUserRefs"], 
        classification["warningsByUserRef"],
        userInfoByUserRef,
        st3_081ByUserRef,
        type3_081["_total"],
        keyColsOnly = deidentify
    )

    mu += webReportNonProxyMachineUsers(
        classification["activeNonProxyMachineUserRefs"], 
        classification["warningsByUserRef"],
        userInfoByUserRef,
        type3_081["_total"],
        st3_081ByUserRef,
        keyColsOnly = deidentify
    )

    mu += webReportRemoteUsers(
        classification["activeRemoteUserRefs"],
        classification["warningsByUserRef"],
        len(classification["activeUserRefs"]), 
        userInfoByUserRef, 
        type3_081["_total"], 
        st3_081ByUserRef,
        classification["remoteExcludeReasonCount"], 
        deidentify,
        keyColsOnly = deidentify
    )
        
    # NOTE: far too crude now - need breaks
    mu += webReportLocalUsers(
        classification["activeLocalUserRefs"], 
        classification["warningsByUserRef"],
        len(classification["activeUserRefs"]), 
        userInfoByUserRef, 
        type3_081["_total"], 
        st3_081ByUserRef,
        deidentify,
        keyColsOnly = deidentify
    )

    userSiteDir = SITE_DIR_TEMPL.format(stationNo)
    reportFile = userSiteDir + ("user.md" if not deidentify else "userDeId.md")
    print("Writing report {}".format(reportFile))
    open(reportFile, "w").write(mu)

"""
Proxy Users

Classifier enforces CONNECTOR PROXY, AV presence and LOA 2

TO ADD:
- workstation_name pattern (ie/ form of TCP Connect)
- forced_close on SO %

Refs: <----- TODO: make explicit in report
https://github.com/vistadataproject/RPCDefinitionToolkit/issues/44
- EDIS - Emergency Department Integration Software (EDIS) https://www.va.gov/vdl/documents/Clinical/Emergency_Dept_Integration_Software/edp_2_1_1_tm.pdf (SPOK: CONNECTOR,EDIS?)
- AVS may be After Visit Summary
- The VistALink Connector Proxy User is an entry on your VistA NEW PERSON file that the PATS application and other web-based applications use to connect to your VistA site. For the PATS application, there will be one data center located in Falling Waters, VA and a second fail-over data center in Hines, IL. A VistALink connector proxy user needs to be set up on your VistA server for the Falling Waters data center and also for the Hines data center.
  (SPOK: VISTALINK,EMC HINES)
- RTLS: "application proxy user 'VIAASERVICE,RTLS APPLICATION PROXY' will be created automatically."
- VPS: 
  (SPOK: CONNECT,VPS)
- Fee Basis Claims System (FBCS) application
- CPGATEWAY,USER: The CP Gateway Service is composed of two subsystems ... via the RPC Broker to retrieve the HL7 message ... Vendor CIS for Clinical Procedures and VistA

Note: tied up with vistalink and two step of connector (with station number
lookup) and then switch to local user <------------ see if two step means two sign ons?
"""
def webReportProxyUsers(activeProxyUserRefs, warningsByUserRef, userInfoByUserRef, st3_081ByUserRef, totalSignOns, keyColsOnly=False):

    totalProxySignOnCount = sum(st3_081ByUserRef[userRef]["_total"] for userRef
 in activeProxyUserRefs)
    mu = """## Proxy Users
    
There are <span class='yellowIt'>{:,}</span> active __Proxy Machine Users__ (user_class is \"CONNECTOR PROXY\") with <span class='yellowIt'>{}</span> signons. All user records have _access_, _verify_ and lack a social while their signons have _LOA_ 2 and don't have \"remote_...\" properties (ie/ CPRS-like combo). These signons probably happen over __VistALink__ and not the plain RPC Broker ...\n\n""".format(len(activeProxyUserRefs), reportAbsAndPercent(totalProxySignOnCount, totalSignOns))
    cols = [":Name [IEN]", "Entered", "SignOns", "Period", "\# IPs"] if keyColsOnly else [":Name [IEN]", "Entered", ":PMO", ":SMOs", ":Keys", "SignOns", "Period", "\# IPs", "Duration", ":Unexpected"] 
    tbl = MarkdownTable(cols)
    allSTs = []
    for userRef in sorted(activeProxyUserRefs, key=lambda x: st3_081ByUserRef[x]["_total"], reverse=True):
        userInfo = userInfoByUserRef[userRef] 
        st = st3_081ByUserRef[userRef]
        allSTs.append(st)
        pmoMU, smosMU, keysMU = muOptionsNKeys(userInfo)
        if "duration" in st:
            if st["_total"] > 1:
                kstatsDur = keyStats(
                    flattenFrequencyDistribution(st["duration"]["byValueCount"])
                )
                durMU = "{}/{}/{}".format(muSeconds(kstatsDur["median"]), muSeconds(kstatsDur["min"]), muSeconds(kstatsDur["max"]))
            else:
                durMU = muSeconds(singleValue(st, "duration"))
        else:
            durMU = ""
        unexpectedMU = "" if userRef not in warningsByUserRef else "/ ".join(warningsByUserRef[userRef])
        if keyColsOnly:
            row = [
                muUserRef(userRef),
                userInfo["date_entered"] if "date_entered" in userInfo else "", 
                reportAbsAndPercent(st["_total"], totalProxySignOnCount),
                muSignOnPeriod(st),
                muBVC(st["ipv4_address"]["byValueCount"], countOnlyIfOver=5) 
            ]
        else:           
            row = [
                muUserRef(userRef),
                userInfo["date_entered"] if "date_entered" in userInfo else "", 
                pmoMU,
                smosMU,
                keysMU,
                reportAbsAndPercent(st["_total"], totalProxySignOnCount),
                muSignOnPeriod(st),
                muBVC(st["ipv4_address"]["byValueCount"], countOnlyIfOver=5),
                durMU,
                unexpectedMU
            ]
        tbl.addRow(row)
    mu += tbl.md() + "\n\n"
            
    return mu
    
"""
Off Key Words BUT Not the DoD User (| Postmaster) 

Expects: machine SSN, visited_from, NO remote_app, LOA 2 (usually), no PMO or keys

Note:
- no PMO or Keys as none specified (in VCB)
- not enforcing LOA 2 as see 200, 2001 combos where first is 1 and then move to 2
- showing remote_user_ien as apparently fixed
- NOT enforcing no remote_app as one CVIX has VISTAWEB login in VCB

CVIX remote_user_ien seems to have a fixed IEN
CVIX_MHVUSER_SSNIEN = "200:412864" # expect 2001 too and 2006_95's >> sign ons
CVIX_USER_SSNIEN = "200:217122" # expect 2001 too; 2006_95 << sign ons 
...
and fixed IP and LOA 1 usually
"""
def webReportNonProxyMachineUsers(activeNonProxyMachineUserRefs, warningsByUserRef, userInfoByUserRef, totalSignOns, st3_081ByUserRef, keyColsOnly=False):

    totalNonProxyMachineSignOnCount = sum(st3_081ByUserRef[userRef]["_total"] for userRef in activeNonProxyMachineUserRefs)
    mu = """## (Non Proxy) Machine Users
    
Besides the _DoD User_, there are <span class='yellowIt'>{:,}</span> active __Non-Proxy Machine Users__ with <span class='yellowIt'>{}</span> signons. These users appear in most VistAs under fabricated social security numbers ...\n\n""".format(
        len(activeNonProxyMachineUserRefs), 
        reportAbsAndPercent(totalNonProxyMachineSignOnCount, totalSignOns)
    )

    # To add: workstation_name - take first part? ipv4_address    
    cols = [":Name [IEN]", "Entered", "SSN", "SignOns", "Period", "Remote Station Id(s)", "Remote IEN(s)", ":IPs"] if keyColsOnly else [":Name [IEN]", "Entered", "SSN", ":SMOs", "SignOns", "Period", "Remote Station Id(s)", "Remote IEN(s)", ":IPs", "Duration", ":Unexpected"]
    tbl = MarkdownTable(cols)
    for userRef in sorted(activeNonProxyMachineUserRefs, key=lambda x: st3_081ByUserRef[x]["_total"], reverse=True):
        userInfo = userInfoByUserRef[userRef]
        st = st3_081ByUserRef[userRef]
        pmoMU, smosMU, keysMU = muOptionsNKeys(userInfo)
        unexpectedMU = "" if userRef not in warningsByUserRef else "/ ".join(warningsByUserRef[userRef])
        if "remote_user_ien" in st:
            if len(st["remote_user_ien"]["byValueCount"].keys()) > 5:
                remoteIENsMU = "_#{:,}_".format(len(st["remote_user_ien"]["byValueCount"].keys()))
            else:
                remoteIENsMU = "/".join(st["remote_user_ien"]["byValueCount"].keys())
        else:
            remoteIENsMU = ""
        if "duration" in st:
            if st["_total"] > 1:
                kstatsDur = keyStats(
                    flattenFrequencyDistribution(st["duration"]["byValueCount"])
                )
                durMU = "{}/{}/{}".format(muSeconds(kstatsDur["median"]), muSeconds(kstatsDur["min"]), muSeconds(kstatsDur["max"]))
            else:
                durMU = muSeconds(singleValue(st, "duration"))
        else:
            durMU = ""
        if keyColsOnly:
            row = [
                muUserRef(userRef),
                userInfo["date_entered"] if "date_entered" in userInfo else "",
                "__{}__".format(userInfo["ssn"]) if "ssn" in userInfo else "",
                reportAbsAndPercent(st["_total"], totalNonProxyMachineSignOnCount),
                muSignOnPeriod(st),
                # NO remote app?
                "/".join(st["remote_station_id"]["byValueCount"].keys()) if "remote_station_id" in st else "",
                remoteIENsMU,
                muBVC(st["ipv4_address"]["byValueCount"], countOnlyIfOver=5)            
            ]
        else:
            row = [
                muUserRef(userRef),
                userInfo["date_entered"] if "date_entered" in userInfo else "",
                "__{}__".format(userInfo["ssn"]) if "ssn" in userInfo else "",
                smosMU,
                reportAbsAndPercent(st["_total"], totalNonProxyMachineSignOnCount),
                muSignOnPeriod(st),
                # NO remote app?
                "/".join(st["remote_station_id"]["byValueCount"].keys()) if "remote_station_id" in st else "",
                remoteIENsMU,
                muBVC(st["ipv4_address"]["byValueCount"], countOnlyIfOver=5),
                durMU,
                unexpectedMU      
            ]
        tbl.addRow(row)
    mu += tbl.md() + "\n\n"
        
    return mu
    
"""
FIRST: exceed and nix
- https://github.com/vistadataproject/RPCDefinitionToolkit/blob/master/Reporters/Users/reportRemoteUsersE.py
- bseEntries = [entry for entry in entries if "remote_app" in entry] etc in
https://github.com/vistadataproject/DataExtractNSync/blob/master/RPCSessionTests/reportUsersAndLogins.py

TODO: needs more
- JLV vs other - see IPs in fmrepo's util
  ... JLV with DoD Ids
  ie/ DoD JLV
  ie/ may show REAL dod ids => de-identify
- non station id/ien combo (need from custom run on SO!) ie/ X:1 in particular
"""
def webReportDoDUser(userRef, userInfo, totalSignons, st, warnings, deidentify):
        
    mu = """## \"DoD User\"
    
One special non proxy, machine user, the __\"DoD User\"__ is used for JLV DoD access and for access by a number of other applications ...\n\n"""
    
    tbl = MarkdownTable([":Property", ":Value"], includeNo=False)
    
    tbl.addRow(["Name \[IEN\]", muUserRef(userRef)])
    tbl.addRow(["Date Entered", userInfo["date_entered"] if "date_entered" in userInfo else ""])
    tbl.addRow(["SSN", "__{}__".format(userInfo["ssn"])])
    pmoMU, smosMU, keysMU = muOptionsNKeys(userInfo)
    tbl.addRow(["SMOs", smosMU])
    tbl.addRow(["Sign ons", reportAbsAndPercent(st["_total"], totalSignons)])
    tbl.addRow(["Sign on period", muSignOnPeriod(st)])
    wdCntr = expandByWeekDay(st)
    tbl.addRow(["Days", ", ".join(["{} [{}]".format("__{}__".format(day) if i < 5 else day, wdCntr[day]) for i, day in enumerate(wdCntr)])])
    noRemoteStationIds = len(st["remote_station_id"]["byValueCount"])
    tbl.addRow(["Station Ids", noRemoteStationIds])
    def topRemoteStationIds(st):
        orderedTopCounts = OrderedDict()
        for i, rsid in enumerate(sorted(st["remote_station_id"]["byValueCount"], key=lambda x: st["remote_station_id"]["byValueCount"][x], reverse=True), 1):
            if i > 5:
                break
            orderedTopCounts[rsid] = st["remote_station_id"]["byValueCount"][rsid]
        return orderedTopCounts
    tbl.addRow(["Top Station Ids", muBVC(topRemoteStationIds(st))])
    fiveOrMoreRemoteStationIds = sum(1 for rsid in st["remote_station_id"]["byValueCount"] if re.match(r'\d\d\d\d\d', rsid))
    tbl.addRow(["5 digit plus Station Ids (DoD?)", fiveOrMoreRemoteStationIds])
    threeAlphaRemoteStationIds = dict((rsid, st["remote_station_id"]["byValueCount"][rsid]) for rsid in st["remote_station_id"]["byValueCount"] if re.match(r'\d\d\d\d?[A-Z]*[A-Z]*$', rsid))
    tbl.addRow(["3 digit [2 alpha] Station Ids (VA)", muBVC(threeAlphaRemoteStationIds, countOnlyIfOver=10)])
    tbl.addRow(["IPs", len(st["ipv4_address"]["byValueCount"])])
    tbl.addRow(["Divisions", muBVC(st["division"]["byValueCount"])])
    if len(warnings):
        tbl.addRow(["Unexpected", "/ ".join(warnings)])
        
    mu += tbl.md() + "\n\n"
    return mu
        
def webReportPostmaster(userRef, userInfo, totalSignons, st, warnings):

    mu = """## Postmaster

Every VistA has __Postmaster__, (one of) the first user in the system ...

"""
    
    tbl = MarkdownTable([":Property", ":Value"], includeNo=False)
    
    tbl.addRow(["Name \[IEN\]", muUserRef(userRef)])
    tbl.addRow(["Date Entered", userInfo["date_entered"] if "date_entered" in userInfo else ""])
    pmoMU, smosMU, keysMU = muOptionsNKeys(userInfo)
    if pmoMU:
        tbl.addRow(["PMO", pmoMU])
    if smosMU:
        tbl.addRow(["SMOs", smosMU])
    if keysMU:
        tbl.addRow(["Keys", keysMU])
    # Division is hardly ever set and then to main site - ignoring
    
    tbl.addRow(["Sign ons", reportAbsAndPercent(st["_total"], totalSignons)])
    tbl.addRow(["Sign on period", muSignOnPeriod(st)])
    
    if len(warnings):
        tbl.addRow(["Unexpected", "/ ".join(warnings)])
    
    mu += tbl.md() + "\n\n"
    return mu
    
"""
REM: the dynamic sniffing of a session would yield much of this (user from certain IPs, remote station id, context chosen is VPR or , workstation 10 etc)

TODO: (review the issues in TK too)
- should add DIVISION restriction (only def div) to REMOTE CHECK
- JLV subset (add to classifier)
  - workstation 10 ie/ the JLV Sub
  - more on VPR SMO ... split em out
    ie mu += "{:,} VPR Remote Users; {:,} no remote_app remote users; {:,} both; {:,} VPR only; {:,} no remote_app only
    "ipv4_addresses": set(JLV_INVA_IPS).union(set(JLV_INVA_EXTRA_IPS)),
            "level_of_assurance": "1",
            "workstation_name": "10"
    and CC JLV
    (besides DoD JLV)
        ipt = IPTracker(defn["name"])
            for so in sos:
                ipt.processSignOn(so)
                globalIPTracker.processSignOn(so) # overalls!
  <-------- VPR subset and tie to JLV IPs too
- REMOTE APP SPEC => BSE subset 
- by APP ie/ CAPRI et al + ENFORCE [classifier] on loa 1/2 (3 for local?) ie/ break JLV vs others in separate tables (fuller report)
- manilla called out; 200 too?
fmrepo/user/catag (move over)
"""
def webReportRemoteUsers(activeRemoteUserRefs, warningsByUserRef, totalActiveUsers, userInfoByUserRef, totalSignOns, st3_081ByUserRef, remoteExcludeReasonCount, deidentify, keyColsOnly=False):

    remoteSignOnCountsByUserRef = dict((userRef, st3_081ByUserRef[userRef]["_total"]) for userRef in activeRemoteUserRefs)
    totalRemoteSignOns = sum(remoteSignOnCountsByUserRef[userRef] for userRef in remoteSignOnCountsByUserRef)
    kstats = keyStats(remoteSignOnCountsByUserRef.values())

    print("About to combine lot's of remote sts (takes time) ...")
    comboST = combineSubTypes([st3_081ByUserRef[userRef] for userRef in activeRemoteUserRefs], forceCountProps=["remote_station_id", "remote_app", "division"])[0]
    print ("... end combo'ing")
    
    mu = """## Remote Users
    
Remote users dominate in every VistA - <span class='yellowIt'>{}</span> - but account for less sign ons - <span class='yellowIt'>{}</span> - than their numbers suggest. The median number of sign ons per remote user is <span class='yellowIt'>{:,}</span>. 
 
""".format(
        reportAbsAndPercent(len(activeRemoteUserRefs), totalActiveUsers),
        reportAbsAndPercent(totalRemoteSignOns, totalSignOns),
        roundFloat(kstats["median"])
    )
    
    remoteStationBVC = comboST["remote_station_id"]["byValueCount"]
    mu += "Remote sign on comes from <span class='yellowIt'>{:,}</span> Remote Stations or <span class='yellowIt'>{:,}</span> three digit stations. The top 10 are ...\n\n".format(
        len(remoteStationBVC),
        len(set(stId[0:3] for stId in remoteStationBVC))
    )    
    tbl = MarkdownTable([":Remote Station Id", "Count of Remote SignOns"])
    for i, stationId in enumerate(sorted(remoteStationBVC, key=lambda x: remoteStationBVC[x], reverse=True), 1):
        if i > 10:
            break
        stationMU = "__{}__".format(stationId)
        if stationId == "459": # Tripler
            stationMU = "__459 [Tripler/JLV Test]__"
        tbl.addRow([
            stationMU, 
            reportAbsAndPercent(remoteStationBVC[stationId], totalRemoteSignOns)
        ])
    mu += tbl.md() + "\n\n"
    
    mu += "Remote signon by week day ...\n\n"
    mu += tblByWeekDay(comboST) + "\n\n"
            
    def muRemoteIds(st): # points to need for custom combine of sno:ien
        stationIdVC = st["remote_station_id"]["byValueCount"]
        # Could happen but ? ... TODO: force IEN count 
        if "byValueCount" not in st["remote_user_ien"]:
            return ""
        ienVC = st["remote_user_ien"]["byValueCount"]
        def muTripler(sid):
            if sid == "459":
                return "__459__"
            return sid 
        if len(stationIdVC) == 1:
            return "{}:{}".format(
                muTripler(stationIdVC.keys()[0]), 
                ienVC.keys()[0]
            )
        if len(ienVC) == 1:
            return "{}:{}".format(
                "/".join([muTripler(sid) for sid in stationIdVC.keys()]), 
                ienVC.keys()[0]
            )
        # TODO: match counts in both to assemble id
        return ""
    mu += "The following shows the top 50 Remote Users ...\n\n"
    cols = [":Name [IEN]", "Entry", "Remote Id(s)", "SignOns", "Period", "Days", ":Remote Apps"] if keyColsOnly else [":Name [IEN]", "Entry", "Remote Id(s)", "SignOns", "Period", "Days", ":Remote Apps", ":Options", "Duration", ":Unexpected"]   
    tbl = MarkdownTable(cols)
    for i, userRef in enumerate(sorted(activeRemoteUserRefs, key=lambda x: st3_081ByUserRef[x]["_total"], reverse=True), 1):
        if i > 50:
            break
        userInfo = userInfoByUserRef[userRef] 
        st = st3_081ByUserRef[userRef]
        # May put VPR in bold later as JLV indicator
        pmoMU, smosMU, keysMU = muOptionsNKeys(userInfo)
        remote_app_count = st["remote_app"]["byValueCount"] if "remote_app" in st else {}
        no_remote_app_count = st["_total"] if "remote_app" not in st else st["_total"] - st["remote_app"]["count"]
        if no_remote_app_count:
            remote_app_count["UNIDENTIFIED"] = no_remote_app_count
        remoteAppMU = ", ".join(["{} ({})".format(remote_app.split(" [")[0], remote_app_count[remote_app]) for remote_app in sorted(remote_app_count, key=lambda x: remote_app_count[x], reverse=True)])
        wdCntr = expandByWeekDay(st)
        if "duration" in st:
            if st["_total"] > 1:
                kstatsDur = keyStats(
                    flattenFrequencyDistribution(st["duration"]["byValueCount"])
                )
                durMU = "{}/{}/{}".format(muSeconds(kstatsDur["median"]), muSeconds(kstatsDur["min"]), muSeconds(kstatsDur["max"]))
            else:
                durMU = muSeconds(singleValue(st, "duration"))
        else:
            durMU = ""
        if keyColsOnly:
            row = [ 
                muUserRef(userRef, ssn="NO SSN" if "ssn" not in userInfo else userInfo["ssn"], deidentify=deidentify),
                userInfo["date_entered"] if "date_entered" in userInfo else "",
                muRemoteIds(st),
                reportAbsAndPercent(st["_total"], totalRemoteSignOns),
                muSignOnPeriod(st),
            ", ".join(["{} [{}]".format("__{}__".format(day) if i < 5 else day, wdCntr[day]) for i, day in enumerate(wdCntr)]),
                remoteAppMU            
            ]
        else:
            row = [
                muUserRef(userRef, ssn="NO SSN" if "ssn" not in userInfo else userInfo["ssn"], deidentify=deidentify),
                userInfo["date_entered"] if "date_entered" in userInfo else "",
                muRemoteIds(st),
                reportAbsAndPercent(st["_total"], totalRemoteSignOns),
                muSignOnPeriod(st),
            ", ".join(["{} [{}]".format("__{}__".format(day) if i < 5 else day, wdCntr[day]) for i, day in enumerate(wdCntr)]),
                remoteAppMU,
                smosMU,
                durMU,
                "" if userRef not in warningsByUserRef else "/ ".join(warningsByUserRef[userRef])
            ]
        tbl.addRow(row)
    mu += tbl.md() + "\n\n"
        
    return mu
    
"""
TODO: 
- broad: FAR TOO catch all -- why does PUG have so many more in Pug than ALake? 
  - and device: ssh vs other
- exclude from remote on MAND_PROPS and ALLOWED_PROPS as opposed to 0's ie/ 0's in here PLUS station_no in locals
"""
def webReportLocalUsers(activeLocalUserRefs, warningsByUserRef, totalActiveUsers, userInfoByUserRef, totalSignOns, st3_081ByUserRef, deidentify, keyColsOnly=False):

    totalLocalSignOns = sum(st3_081ByUserRef[userRef]["_total"] for userRef in activeLocalUserRefs)
    mu = """## Local Users
    
There are <span class='yellowIt'>{}</span> active Local Users with <span class='yellowIt'>{}</span> signons.\n\n""".format(reportAbsAndPercent(len(activeLocalUserRefs), totalActiveUsers), reportAbsAndPercent(totalLocalSignOns, totalSignOns))

    comboST = combineSubTypes([st3_081ByUserRef[userRef] for userRef in activeLocalUserRefs], forceCountProps=["division"])[0]
    mu += "Local users select <span class='yellowIt'>{:,}</span> divisions ...\n\n".format(len(comboST["division"]["byValueCount"]))
    tbl = MarkdownTable([":Division", "Count %"])
    for divisionRef in sorted(comboST["division"]["byValueCount"], key=lambda x: comboST["division"]["byValueCount"][x], reverse=True):
        tbl.addRow([
            "__{}__".format(re.sub(r'4\-', '', divisionRef)), 
            reportAbsAndPercent(comboST["division"]["byValueCount"][divisionRef], totalLocalSignOns)
        ])
    mu += tbl.md() + "\n\n"
    
    mu += "And don't just sign on through device 0 ...\n\n"
    tbl = MarkdownTable([":Device", "Count %"])
    for device in sorted(comboST["device"]["byValueCount"], key=lambda x: comboST["device"]["byValueCount"][x], reverse=True):
        tbl.addRow([
            "__{}__".format(device), 
            reportAbsAndPercent(comboST["device"]["byValueCount"][device], comboST["_total"])
        ])
    mu += tbl.md() + "\n\n"
    
    # TODO: will break on these 
    mu += "And have multiple levels of assurance ...\n\n"
    tbl = MarkdownTable([":LOA", "Count %"])
    for loa in sorted(comboST["level_of_assurance"]["byValueCount"], key=lambda x: comboST["level_of_assurance"]["byValueCount"][x], reverse=True):
        tbl.addRow([
            "__{}__".format(loa), 
            reportAbsAndPercent(comboST["level_of_assurance"]["byValueCount"][loa], comboST["_total"])
        ])
    mu += tbl.md() + "\n\n"
    
    SUPERUSER_KEYS = ["XUMGR"] # removing XUPROG
    superUserRefs = set(userRef for userRef in activeLocalUserRefs if "keys" in userInfoByUserRef[userRef] and len(set(SUPERUSER_KEYS).intersection(set(userInfoByUserRef[userRef]["keys"]))))
    mu += "<span class='yellowIt'>{:,}</span> Local Users are __Superusers__ (those with key {}) ...\n\n".format(len(superUserRefs), "|".join(SUPERUSER_KEYS))
    cols = [":Name [IEN]", "Entry", ":Title", "SignOns", "Period", "Days"] if keyColsOnly else [":Name [IEN]", "Entry", ":Title", "SignOns", "Period", "Days", ":PMO", ":SMOs", ":Keys", ":Unexpected"]
    tbl = MarkdownTable(cols)
    for userRef in sorted(superUserRefs, key=lambda x: st3_081ByUserRef[x]["_total"], reverse=True):
        userInfo = userInfoByUserRef[userRef]
        st = st3_081ByUserRef[userRef]
        pmoMU, smosMU, keysMU = muOptionsNKeys(userInfo)
        wdCntr = expandByWeekDay(st)
        if keyColsOnly:
            row = [
                muUserRef(userRef, ssn="NO SSN" if "ssn" not in userInfo else userInfo["ssn"], deidentify=deidentify),
                userInfo["date_entered"] if "date_entered" in userInfo else "",
                userInfo["title"] if "title" in userInfo else "",
                reportAbsAndPercent(st["_total"], totalLocalSignOns),
                muSignOnPeriod(st),
                ", ".join(["{} [{}]".format("__{}__".format(day) if i < 5 else day, wdCntr[day]) for i, day in enumerate(wdCntr)])            
            ]
        else:
            row = [
                muUserRef(userRef, ssn="NO SSN" if "ssn" not in userInfo else userInfo["ssn"], deidentify=deidentify),
                userInfo["date_entered"] if "date_entered" in userInfo else "",
                userInfo["title"] if "title" in userInfo else "",
                reportAbsAndPercent(st["_total"], totalLocalSignOns),
                muSignOnPeriod(st),
                ", ".join(["{} [{}]".format("__{}__".format(day) if i < 5 else day, wdCntr[day]) for i, day in enumerate(wdCntr)]),
                pmoMU,
                smosMU,
                keysMU,
                "" if userRef not in warningsByUserRef else "/ ".join(warningsByUserRef[userRef])
            ]
        tbl.addRow(row)
    mu += tbl.md() + "\n\n"
        
    return mu
       
# TODO: change to calc length formally in case of gaps 
def muSignOnPeriod(st):
    if len(st["date_time"]["byValueCount"]) == 13:
        soPeriodMU = "EVERY MONTH"
    elif st["date_time"]["firstCreateDate"].split("T")[0] == st["date_time"]["lastCreateDate"].split("T")[0]:
        soPeriodMU = st["date_time"]["lastCreateDate"].split("T")[0]
    elif re.search(r'\-', st["date_time"]["byValueCount"].keys()[0]): # months
        soPeriodMU = "{} - {} ({})".format(st["date_time"]["firstCreateDate"].split("T")[0], st["date_time"]["lastCreateDate"].split("T")[0], len(st["date_time"]["byValueCount"]))
    else:      
        soPeriodMU = "{} - {}".format(st["date_time"]["firstCreateDate"].split("T")[0], st["date_time"]["lastCreateDate"].split("T")[0])
    return soPeriodMU
    
"""
TODO: change -- move to reducing non RPC options 
"""
def muOptionsNKeys(userInfo):
    # rosByLabel = dict((res["label"], res) for res in rpcOptions(stationNo))
    pmoMU = ""
    if "primary_menu_option" in userInfo:
        pmoMU = userInfo["primary_menu_option"]
        # if userInfo["primary_menu_option"] not in rosByLabel:
        #    pmoMU += " [NOT RPC]"
    smosMU = ""
    if "secondary_menu_options" in userInfo:
        # smosMU = ", ".join(sorted([smo if smo in rosByLabel else "{} [NOT RPC]".format(smo) for smo in userInfo["secondary_menu_options"]]))
        if len(userInfo["secondary_menu_options"]) <= 5:
            smosMU = ", ".join(sorted(userInfo["secondary_menu_options"])[0:5])
        else:
            smosMU = "_{:,}_".format(len(userInfo["secondary_menu_options"]))
    keysMU = ""
    if "keys" in userInfo:
        if len(userInfo["keys"]) <= 5:
            keysMU = ", ".join([key for key in sorted(userInfo["keys"])[0:5]])
        else:
            keysMU = "_{:,}_".format(len(userInfo["keys"]))
    return pmoMU, smosMU, keysMU
    
def muUserRef(userRef, ssn=None, deidentify=False):
    name = userRef.split(" [200-")[0]
    if deidentify:
        name = re.sub(r'[A-Za-z]', "X", name)
        if ssn:
            ssn = "XXXXXXXXX"
    ien = userRef.split("[200-")[1][:-1]
    return "__{}__ [{}]".format(
            name,
            "{}/{}".format(ien, ssn) if ssn != None else ien
    )
    
def tblByWeekDay(st):
    ocntr = expandByWeekDay(st, fullDays=True)
    tbl = MarkdownTable([":Day", "Total", "Average"], includeNo=False)
    for i, day in enumerate(ocntr):
        avg = ocntr[day].split("/")[1]
        avgMU = "" if avg == "0" else avg
        tbl.addRow([
            "__{}__".format(day) if i < 5 else day,
            ocntr[day].split("/")[0],
            avgMU
        ])              
    return tbl.md() 
    
"""
Want average as well as total for period - but note that average may be bad
indicator if median is low. Best st can offer.
"""
def expandByWeekDay(st, fullDays=False): 
    def countDaysInPeriod(s, e):
        d1 = datetime.strptime(s.split("T")[0], "%Y-%m-%d")
        d2 = datetime.strptime(e.split("T")[0], "%Y-%m-%d")
        cnter = Counter()
        for d_ord in range(d1.toordinal(), d2.toordinal() + 1):
            d = date.fromordinal(d_ord)
            cnter[d.weekday()] += 1
        return cnter
    createDatePropInfo = st[st["_createDateProp"]]
    dayCnter = countDaysInPeriod(createDatePropInfo["firstCreateDate"], createDatePropInfo["lastCreateDate"])
    weekdays=["Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday", "Sunday"] if fullDays else ["Mon", "Tues", "Wed", "Thurs", "Fri", "Sat", "Sun"]
    cntr = OrderedDict()
    byWeekDay = st["date_time"]["byWeekDay"]
    for wd in sorted(byWeekDay, key=lambda x: int(x)):
        if wd not in byWeekDay:
            continue
        avg = int(round(float(byWeekDay[wd])/ float(dayCnter[int(wd)])))
        val = "{:,}/{:,}".format(byWeekDay[wd], avg)
        cntr[weekdays[int(wd)]] = val
    return cntr
    
def muSeconds(seconds):
    (days, remainder) = divmod(seconds, 86400)
    (hours, remainder) = divmod(remainder, 3600)
    (minutes, seconds) = divmod(remainder, 60)
    if days:
        return "{} {}:{}:{}".format(
            int(days),
            int(hours),
            int(minutes),
            int(seconds)
        )
    if hours:
        return "{}:{}:{}".format(
            int(hours),
            int(minutes),
            int(seconds)
        )
    if minutes:
        return "{}:{}".format(
            int(minutes),
            int(seconds)
        )
    return int(seconds)
            
# ################################# DRIVER #######################
               
def main():

    assert(sys.version_info >= (2,7))
    
    if len(sys.argv) < 2:
        print "need to specify station # ex/ 442 - exiting"
        return
        
    stationNo = sys.argv[1]
    
    userSiteDir = SITE_DIR_TEMPL.format(stationNo)
    if not os.path.isdir(userSiteDir):
        raise Exception("Expect User Site to already exist with its basic contents")
    
    if len(sys.argv) == 3:
        deidentify = sys.argv[2]
    else:
        deidentify = False
        
    if len(sys.argv) == 4:
        smallSOSet = True
    else:
        smallSOSet = False
    
    webReportUser(stationNo, deidentify, smallSOSet)
    
if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf8 -*-

import sys
import os
import re
import json
from collections import defaultdict, Counter
from datetime import datetime

from fmqlutils.reporter.reportUtils import MarkdownTable, reportPercent, reportAbsAndPercent
from fmqlutils.typer.reduceTypeUtils import splitTypeDatas, checkDataPresent, singleValue
from ..webReportUtils import filteredResultIterator, TOP_MD_TEMPL, SITE_DIR_TEMPL, keyStats, roundFloat, reduce200

"""
TOFIX:
- REDO: break must be owner_site too (see if mand)
- local and national difference
- issue of distinct patient: make note that normally from "ALL" type BUT want for ACTIVE vs NOT ACTIVE and per Flag and even per flag per site

Basic PRF

Requires:
- 3_8 (All)
- 26_11 (All)
- 26_13 (Reduction)
- 26_15 (All)
- 200 (All/Indexed-reduce200)

... others to add

TODO: see other xxrepo notes <----- nix when through

TODO v1.1:
- as doing: introduce singleValue, singleValueCount on prop which enforces singletons ie/ reduces need for .keys()
- assigned are mixed with types in table and not in old reports => split out and fill in assignements ala report
- tie to docs: 8925 ... 26_14

TODO v2:
- 391_91 with sources of Patient! (Patient can have > 1 record) ... may break by 'patient'?

NEW TODO: documents tie in (see types in tables) and Treating Facility List tie in ...

> The PRF Manual says that the file, Treating Facility List, is key for designating a Patient as being owned elsewhere and that two HL7 messages are used to synchronize flags with remote systems, one triggered when a patient is added to a VistA, the other when a national flag assignment is made or edited. Both message communications are tracked in PRF HL7 log files.

... see 391.91 entries (if there).
"""
def webReportPRFs(stationNo):

    allThere, details = checkDataPresent(stationNo, [
        {"fileType": "3_8", "check": "ALL"},
        {"fileType": "26_11", "check": "ALL"},
        {"fileType": "26_13", "check": "TYPE"},
        {"fileType": "26_15", "check": "ALL"},
        {"fileType": "200", "check": "ALL"} # as indexed
    ])
    if not allThere:
        raise Exception("Some required data is missing - {}".format(details))
                
    all26_13Data, st26_13s = splitTypeDatas(stationNo, "26_13", expectSubTypeProperty="status-flag_name-owner_site")
    stActive26_13s = [st26_13 for st26_13 in st26_13s if re.search(r':ACTIVE', st26_13["_subTypeId"])]
    activeFlagUse = dict((st26_13["flag_name"]["byValueCount"].keys()[0], st26_13["_total"]) for st26_13 in stActive26_13s) 
            
    mu = """---
layout: default
title: {} Patient Record Flags
---

""".format(stationNo) 

    mu += """Patient Record Flags (PRFs) are
     
> used to alert Veterans Health Administration (VHA) medical staff and employees of patients whose behavior and characteristics may pose a threat either to their safety, the safety of other patients, themselves, or compromise the delivery of quality health care. PRFs’ assignments are displayed during the patient look-up process

Flag assignment involves document creation and maintenance, use of email and HL7 messaging.

"""

    mu += webReportAssignments(stationNo, all26_13Data, stActive26_13s)

    fmu, mailGroupUse, clsIDocTypeUse, clsIIDocTypeUse = webReportFlagTypes(stationNo, activeFlagUse)
    mu += fmu
    
    mu += webReportMailGroupUse(stationNo, mailGroupUse)
    
    mu += webReportHL7Use(stationNo)
        
    userSiteDir = SITE_DIR_TEMPL.format(stationNo)
    open(userSiteDir + "prfs.md", "w").write(mu)
    
"""
Expand out but also decide interplay of showing assignment #'s above and this. 

TODO: remote #'s from above and just note if used or not AND off 26_14 NEW ASSIGNMENT
dates
"""
def webReportAssignments(stationNo, all26_13Data, stActive26_13s):

    cntNational = sum(st26_13["_total"] for st26_13 in stActive26_13s if re.search(r'\_15', st26_13["_subTypeId"]))
    cntLocal = sum(st26_13["_total"] for st26_13 in stActive26_13s if re.search(r'\_11', st26_13["_subTypeId"]))
    cntUniquePatients = sum(len(st26_13["patient_name"]["byValueCount"]) for st26_13 in stActive26_13s)
    
    mu = """### Flag Assignments
    
This VistA has <span class='yellowIt'>{:,}</span> active and <span class='yellowIt'>{:,}</span> inactive Patient Record Flag (PRF) assignments, <span class='yellowIt'>{:,}</span> in total. There are <span class='yellowIt'>{:,}</span> Class I (Nationally Defined) and <span class='yellowIt'>{:,}</span> Class II (Locally Defined) active flag assignments for <span class='yellowIt'>{:,}</span> patients.

""".format(all26_13Data["status"]["byValueCount"]["1:ACTIVE"], all26_13Data["status"]["byValueCount"]["0:INACTIVE"], all26_13Data["_total"], cntNational, cntLocal, cntUniquePatients)

    return mu

def webReportFlagTypes(stationNo, activeFlagUse):

    mu = "### Flag Types\n\n"

    mailGroupUse = defaultdict(list)
    
    # Nationals 
    tbl = MarkdownTable(["Flag", "Type", "Description", "Document Title", "email Group", "R/N", "Assigned"])
    resourceIter = filteredResultIterator(stationNo, "26_15")
    props = ["name", "type__03", "description", "tiu_pn_title", "review_mail_group", "review_frequency_days", "notification_days"]
    clsIDocTypeUse = defaultdict(int)
    total = 0
    for resource in resourceIter:
        fId = "{} [{}]".format(resource["label"], resource["_id"])
        if resource["status"] != "1:ACTIVE":
            raise Exception("Expected all to be active")
        if len(set(resource.keys()) - (set(["_id", "type", "label", "status"]).union(set(props)))):
            raise Exception("Unexpected property for national flag")
        row = []
        for prop in props:
            if prop == "name":
                if fId in activeFlagUse:
                    labelMU = "__{}__".format(resource["label"]) 
                    flagUseMU = activeFlagUse[fId] 
                else:
                    labelMU = resource["label"]
                    flagUseMU = ""
                row.append(labelMU)
                continue
            if prop not in resource:
                row.append("")
                continue
            if prop == "review_frequency_days":
                rnValue = resource[prop]
                continue
            if prop == "notification_days":
                rnValue += " / " + resource[prop]
                row.append(rnValue)
                continue
            if prop == "type__03":
                row.append(resource[prop]["label"][0])
                continue
            if isinstance(resource[prop], dict):
                row.append("{}".format(resource[prop]["label"]))
            elif re.search(r'\d:', resource[prop]):
                row.append(resource[prop].split(":")[1])
            elif len(row) == 0:
                row.append("__{}__".format(resource[prop]))
            else:
                row.append(re.sub(r'\n', ' ', resource[prop]))
            if prop == "review_mail_group":
                mailGroupUse[resource[prop]["id"]].append(fId)
            if prop == "tiu_pn_title":
                clsIDocTypeUse["{}".format(resource[prop]["id"])] += 1 
        row.append(flagUseMU)
        total += 1
        tbl.addRow(row)
    mu += "Every VistA Systems supports the following <span class='yellowIt'>4</span> __Class I__ (Nationally Defined) flags. The following table lists these flags along with their related document titles and mail email groups and their review and notify (R/N) times (in days) ...\n\n"
    mu += tbl.md() + "\n\n"
        
    # Locals
    tbl = MarkdownTable(["Flag", "Type", "Description", "Document Title", "email Group", "R/N", "Assigned"])
    resourceIter = filteredResultIterator(stationNo, "26_11")
    props = ["name", "type__03", "description", "tiu_pn_title", "review_mail_group", "review_frequency_days", "notification_days"]
    clsIIDocTypeUse = defaultdict(int)
    total = 0
    noAssigned = 0
    rows = []
    for resource in resourceIter:
        if len(set(resource.keys()) - (set(["_id", "type", "label", "status"]).union(set(props)))):
            raise Exception("Unexpected property for national flag")
        total += 1
        if not ("status" in resource and resource["status"] == "1:ACTIVE"):
            continue
        fId = "{} [{}]".format(resource["label"], resource["_id"])
        row = []
        for prop in props:
            if prop == "name":
                if fId in activeFlagUse:
                    labelMU = "__{}__".format(resource["label"]) 
                    flagUseMU = activeFlagUse[fId] 
                    noAssigned += 1
                else:
                    labelMU = resource["label"]
                    flagUseMU = ""
                row.append(labelMU)
                continue
            if prop not in resource:
                row.append("")
                continue
            if prop == "review_frequency_days":
                rnValue = resource[prop]
                continue
            if prop == "notification_days":
                rnValue += " / " + resource[prop]
                row.append(rnValue)
                continue
            if isinstance(resource[prop], dict):
                row.append("{}".format(resource[prop]["label"]))
            elif re.search(r'\d:', resource[prop]):
                row.append(resource[prop].split(":")[1])
            elif len(row) == 0:
                row.append("__{}__".format(resource[prop]))
            else:
                row.append(re.sub(r'\n', ' ', resource[prop]))
            if prop == "review_mail_group":
                mailGroupUse[resource[prop]["id"]].append(fId)
                continue
            if prop == "tiu_pn_title":
                clsIIDocTypeUse["{}".format(resource[prop]["id"])] += 1 
        row.append(flagUseMU)
        rows.append(row)
    for row in sorted(rows, key=lambda x: x[0]):
        tbl.addRow(row)
    mu += "This system defines <span class='yellowIt'>{:,}</span> __Class II__ (Locally Defined) flags, only <span class='yellowIt'>{:,}</span> of which are currently assignable (active) and only <span class='yellowIt'>{:,}</span> of those (in bold below) are used in active flag assignments ...\n\n".format(total, len(rows), noAssigned)    
    mu += tbl.md() + "\n\n"
    
    return mu, mailGroupUse, clsIDocTypeUse, clsIIDocTypeUse
    
"""
owner_site 

> current site that owns this patient flag assignment. Patient assignments may only be edited by the owner site. The owner site normally corresponds to the site providing primary care to the patient.

... may differ from originating_site.

TODO: 
- BUG: unique patients only works if add "owner_site" to the subTypeId
- add in state and station number for sites (need off constants) 
"""
def webReportOwnerSite(stationNo, stActive26_13s):

    totalAssignments = 0
    nationalAssignmentsByOwnerSite = Counter()
    localAssignmentsOfThisSite = 0
    uniquePatientsOfOwnerSite = defaultdict(lambda: set())
    for st26_13 in stActive26_13s:
        if not ("owner_site" in st26_13 and "byValueCount" in st26_13["owner_site"]):
            raise Exception("Expected Reduction of 26_13 to ensure owner_site is present in all actives AND the property be explicitly counted")
        if not ("patient_name" in st26_13 and "byValueCount" in st26_13["patient_name"]):
            raise Exception("Expected Reduction of 26_13 to ensure patient_name is present in all actives AND the property be explicitly counted")
        totalAssignments += st26_13["_total"]
        thisSite = ""
        for ownerSite in st26_13["owner_site"]["byValueCount"]:
            if re.search(r'\_15', st26_13["_subTypeId"]):
                nationalAssignmentsByOwnerSite[ownerSite] += st26_13["owner_site"]["byValueCount"][ownerSite]
                continue
            if thisSite:
                if thisSite != ownerSite:
                    raise Exception("Only expect one site - this one - to have class I nationals as well as local assignments")
            else:
                thisSite = ownerSite
            localAssignmentsOfThisSite += st26_13["_total"] # ie/ all to this site
            for patientName in st26_13["patient_name"]["byValueCount"]: # for now, only local
                uniquePatientsOfOwnerSite[ownerSite].add(patientName)
            
    mu = "### Owner VistAs\n\n"
    
    mu += "Including itself, <span class='yellowIt'>{:,}</span> VistAs contribute Active Flag Assignments to this VistA. The following table lists and ranks VistAs contributing 20 or more assignments or which are VistAs of interest. Not surprisingly, the local VistA contributes the most assignments ...\n\n".format(len(nationalAssignmentsByOwnerSite))
            
    tbl = MarkdownTable(["Rank", "Name", "Flag Count (Class I/ Class II)", "\%", "Unique Patients"])
    tbl.addRow([
        1,
        "__{}__".format(thisSite),
        "{:,} ({:,} / {:,})".format(
            nationalAssignmentsByOwnerSite[thisSite] + localAssignmentsOfThisSite,
            nationalAssignmentsByOwnerSite[thisSite],
            localAssignmentsOfThisSite
        ),
        reportPercent(nationalAssignmentsByOwnerSite[thisSite] + localAssignmentsOfThisSite, totalAssignments),
        len(uniquePatientsOfOwnerSite[thisSite])
    ])
    for i, ownerSite in enumerate(sorted(nationalAssignmentsByOwnerSite, key=lambda x: nationalAssignmentsByOwnerSite[x], reverse=True), 1):
        if ownerSite == thisSite:
            continue
        tbl.addRow([
            i,
            ownerSite,
            "{:,}".format(nationalAssignmentsByOwnerSite[ownerSite]),
            reportPercent(nationalAssignmentsByOwnerSite[ownerSite], totalAssignments),
            len(uniquePatientsOfOwnerSite[ownerSite])
        ])  
        
    mu += tbl.md() + "\n\n"
    
    return mu

"""
MAIL GROUPS
""" 
def webReportMailGroupUse(stationNo, mailGroupUse):   
     
    userInfoById = reduce200(stationNo)
        
    mu = "### email Groups\n\n"

    rows = []
    cols = ["Group", "Description", "Members", "Flag(s)"]
    tbl = MarkdownTable(cols)
    resourceIter = filteredResultIterator(stationNo, "3_8")
    flagsCovered = set()
    for resource in resourceIter:
        if resource["_id"] not in mailGroupUse:
            continue
        members = []
        if "member" in resource:
            for memberInfo in resource["member"]:
                memberRef = "{} [{}]".format(memberInfo["member"]["label"], memberInfo["member"]["id"])
                title = userInfoById[memberRef]["title"] if "title" in userInfoById[memberRef] else title
                members.append("{} ({})".format(memberInfo["member"]["label"], title))
        membersMU = ", ".join(sorted(members))
        row = ["__{}__".format(resource["label"]), re.sub(r'\n', ' ', resource["description"]) if "description" in resource else "", membersMU]
        # Note: by defn of 26.11/15, flag types have only one group each
        flagUsesMUs = ["{}".format(flagRef.split(" [")[0]) if re.search(r'\_11', flagRef) else "__{}__".format(flagRef.split(" [")[0]) for flagRef in sorted(mailGroupUse[resource["_id"]])]
        flagsCovered |= set(mailGroupUse[resource["_id"]])
        row.append(", ".join(flagUsesMUs))
        rows.append(row)
    for row in sorted(rows, key=lambda x: x[0]):
        tbl.addRow(row)
    mu += "<span class='yellowIt'>{:,}</span> mail groups are active for <span class='yellowIt'>{:,}</span> flags, <span class='yellowIt'>{:,}</span> of which, highlighted in bold, are Class I (National) flags ...\n\n".format(len(rows), len(flagsCovered), sum(1 for flagRef in flagsCovered if re.search(r'\_15', flagRef)))
    mu += tbl.md() + "\n\n"
    
    return mu
    
def webReportHL7Use(stationNo):

    # site_transmitted_to ... total by that!
    _17typeDatas, sts = splitTypeDatas(stationNo, "26_17", expectSubTypeProperty="")
    _19typeDatas, sts = splitTypeDatas(stationNo, "26_19", expectSubTypeProperty="")
    
    mu = "## HL7 Messaging for Class I Synchronization\n\n"
    for typeData, msgName in [(_17typeDatas[0], "Update R01"), (_19typeDatas[0], "Query R02")]:

        bvc = typeData["site_transmitted_to"]["byValueCount"]
        mdTable = MarkdownTable(["Site", "SSN", "Transmission Tos"], includeNo=False)
        sepRowIn = False
        for i, site in enumerate(sorted(bvc, key=lambda x: bvc[x], reverse=True), 1):
            if not (i < 11 or re.search(r'\-(663|640|668|442)', site)):
                if i > 10 and not sepRowIn:
                    mdTable.addSeparatorRow()
                    sepRowIn = True
                continue
            mdTable.addRow([
                i,
                site.split(" [")[0] if not re.search(r'\-(663|640|668|442)', site) else "__{}__".format(site.split(" [")[0]),
                re.search(r'\-(\d+)\]', site).group(1),
                reportAbsAndPercent(bvc[site], typeData["_total"])
            ])
        mu += "<span class='yellowIt'>{}</span> Transmissions of {} updates ...\n\n".format(typeData["_total"], msgName)
        mu += mdTable.md() + "\n\n"
        
    """
    The PRF subsystem uses two HL7 messages (Update RO1) and (Query R02) to synchronize Class I (National) flag assignments with other VistAs. Synchronization relies on VistA’s Treating Facility file. Both message types have their own logs in the system. According to this system’s logs ...
        * 6980 Update R01’s have been sent. These were triggered by updates to national flag assignments.
        * 77,446 Query R02’s have been sent. These were triggered by various types of patient registration.  
    """
    
    return mu
    
"""
Document Tie in

    In 26.14
            9. tiu_type_pn_link - POINTER - 759 (17%) - 8925_1
                1. PATIENT RECORD FLAG CATEGORY I - HIGH RISK FOR SUICIDE [8925_1-258] - 374 (49.3%)
                2. PATIENT RECORD FLAG CATEGORY II - HIGH RISK OF SUICIDE [8925_1-2113] - 370 (48.7%)
                3. PATIENT RECORD FLAG CATEGORY II - DISRUPTIVE BEHAVIOR [8925_1-2055] - 8 (1.1%)
                4. PATIENT RECORD FLAG CATEGORY I [8925_1-1402] - 7 (0.9%)
"""
    
# ################################# DRIVER #######################
               
def main():

    assert(sys.version_info >= (2,7))

    if len(sys.argv) < 2:
        print "need to specify station # ex/ 442 - exiting"
        return
        
    stationNo = sys.argv[1]
    
    userSiteDir = SITE_DIR_TEMPL.format(stationNo)
    if not os.path.isdir(userSiteDir):
        raise Exception("Expect User Site to already exist with its basic contents")
    
    webReportPRFs(stationNo)
                 
if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf8 -*-

import sys
import os
import re
import json
from collections import defaultdict, Counter
from datetime import datetime
import numpy

from fmqlutils.reporter.reportUtils import MarkdownTable, reportPercent, reportAbsAndPercent
from fmqlutils.typer.reduceTypeUtils import splitTypeDatas, checkDataPresent, singleValue
from ..webReportUtils import filteredResultIterator, reduce200, TOP_MD_TEMPL, SITE_DIR_TEMPL, keyStats, roundFloat

"""
Images types, references (TIU, Radiology), Acquisition Devices, Users ...

TODO:
- bug: dicom in last table's orphans
- add in more on typing using docs and patient #'s (see GOAL)

Note: may aspects of images will appear in other more specific reports and some of the
breakdowns below (TODO) may become functions to enable that reporting ex/ Image-centric documents etc.

GOAL: move towards patient centric view of images and using reference typing to type images as native form not good enough (expand on this theme). Also expand on user access to images. Overall, WANT TO FOLLOW SAME PATTERNS as other data types. 
"""
    
def webReportImaging(stationNo):

    # Too many here - part of plus
    allThere, details = checkDataPresent(stationNo, [
    
        {"fileType": "702_09", "check": "ALL"}, # in 2005RF too
        {"fileType": "2005", "check": "YR1"}, # object_type-pdf-capture_application (off 2005RF)
        {"fileType": "2005_02", "check": "TYPE"},
        {"fileType": "2005_03", "check": "ALL"},
        {"fileType": "2006_04", "check": "ALL"},
        {"fileType": "2006_532", "check": "ALL"}, # SOP ids - get labels
        {"fileType": "2006_82", "check": "YR1"}, # workstation (used for 2005RF too)
        {"fileType": "8925", "check": "YR1"}, # document_type (used for 2005RF too)

        # Just for 2005 RF
        {"fileType": "74", "check": "ALL"}, # will index in a shelve
        {"fileType": "8925_1", "check": "ALL"},
        {"fileType": "8925_91", "check": "ALL"} # shelve to have doc by img and vica versa

    ])
    if not allThere:
        raise Exception("Some required data is missing - {}".format(details))

    # ###################### Preparation/ Preprocessing #########################
          
    allTypeData, subTypeDatas = splitTypeDatas(stationNo, "2005", reductionLabel="YR1", expectSubTypeProperty="object_type-pdf-capture_application")
    totalAll = allTypeData["_total"]
        
    """
    Enforce things about capture_application and image_save_by etc as interpretation
    and derivation rely on them
    """
    def enforce2005(typeInfo):
    
        """
        Acquisition Device (2006.04): mandatory for I:Import API and D:DICOM Gateway but 
        never C:Capture Workstation 
        """
        if "capture_application" not in typeInfo:
            if "acquisition_device" in typeInfo:
                raise Exception("Didn't expect acquisition device if no capture_application")
        else:
            totalCAID = sum(typeInfo["capture_application"]["byValueCount"][ca] for ca in typeInfo["capture_application"]["byValueCount"] if ca != "C:Capture Workstation")
            if totalCAID and ("acquisition_device" not in typeInfo or typeInfo["acquisition_device"]["count"] != totalCAID):
                raise Exception("Expected acquisition device if capture application is D:DICOM or I:Import")
                
        """
        Image Save By:
        - mandatory for I:Import API and C:Capture and even for DICOM Gateway
        if it leads to DICOM:TIU (unsure how that can happen!) ie/ not for DICOM:RADIOLOGY
          ... possible that DICOM:TIU is one user too??????
        - I:Import API leads to one and only one user across the system ie a default guy!
        """
        totalCAIC = sum(typeInfo["capture_application"]["byValueCount"][ca] for ca in typeInfo["capture_application"]["byValueCount"] if "capture_application" in typeInfo and ca != "D:DICOM Gateway")
        if ("image_save_by" in typeInfo and (typeInfo["image_save_by"]["count"] != totalCAIC)) or (totalCAIC and "image_save_by" not in typeInfo):
            raise Exception("Expected 'image_save_by' to match combo of capture I and C")
        totalCAI = sum(typeInfo["capture_application"]["byValueCount"][ca] for ca in typeInfo["capture_application"]["byValueCount"] if "capture_application" in typeInfo and ca == "I:Import API")
        # in == <=> as typeInfo is fixed to one capture_application
        if "capture_application" in typeInfo and "I:Import API" in typeInfo["capture_application"]["byValueCount"]:
            if "image_save_by" not in typeInfo:
                raise Exception("Import API but no Image Save User Set?")
            noI = typeInfo["capture_application"]["byValueCount"]["I:Import API"]
            possImportAPIUsers = set(userId for userId in typeInfo["image_save_by"]["byValueCount"] if typeInfo["image_save_by"]["byValueCount"][userId] == noI)
            if not len(possImportAPIUsers):
                raise Exception("Expect (at least) one user to a/c for Import API!")
    for subTypeData in subTypeDatas:
        enforce2005(subTypeData)
        
    # ####################### Make Report / Web Page #############################
            
    userSiteDir = SITE_DIR_TEMPL.format(stationNo)
    if not os.path.isdir(userSiteDir):
        raise Exception("Expect User Site to already exist with its basic contents")

    title = "Imaging".format(stationNo)
    mu = TOP_MD_TEMPL.format("Imaging", title)
    mu += """VistA Imaging ...
    
> support(s) the goal making a patient’s complete medical record available online

"""
    
    # ################################ Format / Reference #################

    mu += webReportFormatReference(stationNo, allTypeData, subTypeDatas)
    
    # ################################ Others Plus ##################
    
    """
    The CAP and CAP-IMPORT actions in 2006_82 are already mined into an enhanced
    2005 to allow WS identification for CAPTURE and redundant definition for
    IMPORT API. Bringing this in again, allows more:
    - if only one user in byWS then can id the real user for IMPORT API as all seem 
    set to one
    - activity of READ-ONLY so can see number of workstations involved.
    
    _2006_82AllTypeInfo, _2006_82SubTypeInfos = splitTypeDatas(stationNo, "2006_82", reductionLabel="YR1", expectSubTypeProperty="workstation")
    mu += webReportImageAcquisition(stationNo, subTypeDatas, _2006_82SubTypeInfos)
    """
    # mu += webReportImageDocumentTypes(stationNo) 
    # mu += crudeTIULinkReport(stationNo)
            
    open(userSiteDir + "imaging.md", "w").write(mu)
        
# #################### Format/ Reference ###################

def webReportFormatReference(stationNo, allTypeData, subTypeDatas):

    """
    subtypes - splitting out (XRAY)_GROUP as not 'really' an image ie/ IMAGE GROUP
    - with fileref => individual images
    - without: expect object_group for children and grouper
    """
    def distinguishFileRefAndGroupSubTypes(subTypeDatas):
        filerefSubTypesByOT = defaultdict(list)
        groupSubTypes = []
        for i, subTypeInfo in enumerate(sorted(subTypeDatas, key=lambda x: singleValue(x, "object_type"), reverse=True), 1): 
            ot = singleValue(subTypeInfo, "object_type").split(" [")[0]
            if ot == "XRAY GROUP":
                if "object_group" not in subTypeInfo:
                    raise Exception("Expected ALL groupers to have children")
                groupSubTypes.append(subTypeInfo)
                continue
            if "object_group" in subTypeInfo:
                raise Exception("Expected NO Individual Images to have children")
            filerefSubTypesByOT[ot].append(subTypeInfo)
        return filerefSubTypesByOT, groupSubTypes 
    filerefSubTypesByOT, groupSubTypes = distinguishFileRefAndGroupSubTypes(subTypeDatas)

    mu = ""
        
    """
    Individual Images: all have fileref 
    
    Missing: [TODO]
    - patient as didn't capture byValueCount for 2 in reduction. Must redo OR redo
    uniques iff do subtyping just by type and not type_pdf
    
    Note: keeping 'capture' for now - will expand Acquisition below to report on formats
    and capture mode and take out of here later. TODO
    """
    tbl = MarkdownTable(["Type", "Count", "Form", "Capture", "Direct Referenced", "Group Referenced", "Orphan"])
    imageTotal = 0
    userCntByTypeSubType = defaultdict(lambda: defaultdict(dict))
    allTotalLessDICOM = 0
    allTotalOrphansLessDICOM = 0 
    pdfLabelsSeen = set()  
    importAPIUser = "" 
    
    """
    TO AUDIT: iffy logic - assuming that individual images get link as well as
    parent if parent has link. There is NO separate check that only a parent
    link. Doesn't seem to matter but??
    """
    # Global one to isolate out just by PDFLabel irrespective of image type (ot)
    # ... will see only TIU (and a few NOSET) apply for import api
    byCAByPDFLabel = defaultdict(lambda: defaultdict(int))
    # type ie/ DICOM IMAGE etc
    summs = []
    summPies = []
    for typeId in sorted(filerefSubTypesByOT.keys(), key=lambda x: sum(y["_total"] for y in filerefSubTypesByOT[x]), reverse=True):
        
        subTypeInfos = filerefSubTypesByOT[typeId]
        
        typeTotal = sum(subTypeInfo["_total"] for subTypeInfo in subTypeInfos)
        imageTotal += typeTotal
        if typeId != "DICOM IMAGE":
            allTotalLessDICOM += typeTotal
        
        suffixesCount = defaultdict(int)
        caCount = defaultdict(int)
        oiCount = defaultdict(int)

        linksDirectByPDF = defaultdict(int)
        linksFromParentByPDF = {}
        orphanNo = 0
            
        for subTypeInfo in subTypeInfos:
                                
            # typeLabel = subTypeValue.split("/")[0]
            typeLabel = singleValue(subTypeInfo, "object_type").split(" [")[0]
            pdfLabel = singleValue(subTypeInfo, "pdf", "NOSET")
            pdfLabelsSeen.add(pdfLabel)
                            
            for suffix in subTypeInfo["fileref_suffix"]["byValueCount"]:
                suffixesCount[suffix] += subTypeInfo["fileref_suffix"]["byValueCount"][suffix]
            for ca in subTypeInfo["capture_application"]["byValueCount"]:
                caCount[ca] += subTypeInfo["capture_application"]["byValueCount"][ca]
                byCAByPDFLabel[ca][pdfLabel] += subTypeInfo["capture_application"]["byValueCount"][ca]
                
            # User - for CA == capture_workstation only - come back to him later
            # ... note: some Capture Workstation missing em!
            if "image_save_by" in subTypeInfo:
                for userId in subTypeInfo["image_save_by"]["byValueCount"]:
                    userCntByTypeSubType[typeId][pdfLabel][userId] = subTypeInfo["image_save_by"]["byValueCount"][userId]
            if "origin_index" in subTypeInfo:
                for oi in subTypeInfo["origin_index"]["byValueCount"]:
                    oiCount[oi] += subTypeInfo["origin_index"]["byValueCount"][oi]
                            
            """
            The following leverages:
            - one NOSET vs others
            - others "parent_data_file" < "_total" => parent is responsible!
            """
            if pdfLabel != "NOSET":
                if "parent_data_file" in subTypeInfo:
                    linksDirectByPDF[pdfLabel] += subTypeInfo["parent_data_file"]["count"]
                if "group_parent" in subTypeInfo:
                    rangeCount = subTypeInfo["group_parent"]["rangeCount"] if "rangeCount" in subTypeInfo["group_parent"] else len(subTypeInfo["group_parent"]["byValueCount"])
                    # rem: B3 sub typing so must be incremental
                    if pdfLabel not in linksFromParentByPDF:
                        linksFromParentByPDF[pdfLabel] = [subTypeInfo["group_parent"]["count"], rangeCount]
                    else: 
                        linksFromParentByPDF[pdfLabel][0] += subTypeInfo["group_parent"]["count"]
                        linksFromParentByPDF[pdfLabel][1] += rangeCount
                continue
                
            # From NOSET (must exist as here!) ... can be C or I
            orphanNo += subTypeInfo["_total"]    
            if typeId != "DICOM IMAGE":
                allTotalOrphansLessDICOM += subTypeInfo["_total"]  
            
        summ = {"name": re.sub(r'\_', ' ', typeId)}
        summPie = {"Type": re.sub(r'\_', ' ', typeId)}
            
        # Note: argument to take parent and not direct DICOM which would then show orphans
        # ... seems like RAD goes to GROUP and not individual and may be missing.             
        totalLinked = sum(linksFromParentByPDF[x][0] for x in linksFromParentByPDF if typeId != "DICOM_IMAGE") + sum(linksDirectByPDF[x] for x in linksDirectByPDF)
 
        typeTotalMU = "{:,}".format(typeTotal)
        
        summPie["Count"] = typeTotal
 
        if len(suffixesCount) == 1:
            suffixesMU = suffixesCount.keys()[0]
        else:
            suffixesMU = "<br><br>".join(["{} ({:,})".format(suffix, suffixesCount[suffix]) for suffix in sorted(suffixesCount, key=lambda x: suffixesCount[x], reverse=True)])

        if len(caCount) == 1:
            captureMU = caCount.keys()[0].split(":")[1]
        else:
            captureMU = "<br><br>".join(["{} ({:,})".format(vc.split(":")[1], caCount[vc]) for vc in sorted(caCount, key=lambda x: caCount[x], reverse=True)])

        if len(oiCount) == 1:
            originMU = oiCount.keys()[0].split(":")[1]
        else:
            originMU = "<br><br>".join(["{} ({:,})".format(vc.split(":")[1], oiCount[vc]) for vc in sorted(oiCount, key=lambda x: oiCount[x], reverse=True)])  

        """
        Direct Links
        """
        directLinkMU = "<br><br>".join(["{} ({:,})".format(re.sub(r'_', ' ', pdf), linksDirectByPDF[pdf]) for pdf in linksDirectByPDF])
        
        summ["directReferenced"] = sum(linksDirectByPDF[pdf] for pdf in linksDirectByPDF)        

        """
        Parent/Group
        """
        parentLinkMU = "<br><br>".join(["{} ({:,}/{:,})".format(re.sub(r'_', ' ', gpdf), linksFromParentByPDF[gpdf][0], linksFromParentByPDF[gpdf][1]) for gpdf in linksFromParentByPDF]) 
        
        summ["groupReferenced"] = sum(linksFromParentByPDF[gpdf][0] for gpdf in linksFromParentByPDF) if typeId != "DICOM_IMAGE" else 0 # as redundant
        
        orphansMU = "" if orphanNo == 0 else "{:,}".format(orphanNo)
        
        summ["orphan"] = orphanNo

        # removed 'originMU' as better looked at below. Seems not to be accurate.        
        tbl.addRow([
            "__{}__".format(re.sub(r'\_', ' ', typeId)), 
            typeTotalMU, 
            suffixesMU, 
            captureMU, 
            directLinkMU,
            parentLinkMU, 
            orphansMU
        ])
    
        # Don't include DICOM in summary    
        if not re.match(r'DICOM', typeId):
            summs.append(summ)
        summPies.append(summPie)
                
    objectTypesInfo, sts = splitTypeDatas(stationNo, "2005_02", expectSubTypeProperty="")        
    
    totalXRG = sum(x["_total"] for x in groupSubTypes)
    
    overallFirstCreateDate = allTypeData["_firstCreateDate"].split("T")[0] 
    overallLastCreateDate = allTypeData["_lastCreateDate"].split("T")[0]
    
    mu += """<span class='yellowIt'>{imageTotal:,}</span> individual images were entered into this VistA from {overallFirstCreateDate} through {overallLastCreateDate}, the last full year for which data is available. <span class='yellowIt'>{filerefSubTypesByOTCount:,}</span> formats of image were entered from the <span class='yellowIt'>{objectTypesInfoTotal:,}</span> supported by the _VistA Imaging_ Subsystem. These images are for <span class='yellowIt'>{allTypeDataPatientRangeCount:,}</span> distinct patients.

<span class='yellowIt'>{totalXRG:,}</span> “XRAY” GROUP “Images” were also entered during this period. These group individual images and, despite their name, don’t just group XRAYs or even DICOM Images - they can group PDFs, TIFs and other Image formats too. Grouping allows individual TIU Notes (TIU) and Radiology Notes (RADIOLOGY) reference a group of Images.
    
An image can be entered in one of three ways - _Capture Workstation_ for manually entering images and _DICOM Gateway_ and _Import API_, both automated entry mechanisms.

""".format(
        imageTotal=imageTotal, 
        overallFirstCreateDate=overallFirstCreateDate, 
        overallLastCreateDate=overallLastCreateDate, 
        filerefSubTypesByOTCount=len(filerefSubTypesByOT.keys()), 
        objectTypesInfoTotal=objectTypesInfo["_total"], 
        allTypeDataPatientRangeCount=allTypeData["patient"]["rangeCount"],
        totalXRG=totalXRG
    )
    
    # #################### Formats and References ###################
    """
    Two things - by format, ac dev AND reference pattern. Probably too much
    as both dimensions can be looked at separately.
    
    TO CHANGE:
    ie/ BREAK THIS PROPERLY AFTER DOING AC DEV. Ex/ of what's lost.
    ie/ the dimension of format and ac and ref both there ... two different
    items ie/ format:ref and format:ac
    
    Cheyenne:
    - 67 :C for DICOM ... sub refs to that is TIU (30), TMP (37) both from non main
    institution ie/ kind of shows no gateway there to take in DICOMS so doing in CAPTURE
    (should see those WS's as being from those places!)
    
    Note: devices too - references. Question of whether ac dev and references
    should be in same subsection. Doing AC DEV below and if appropriate then will keep
    """
    
    mu += "## Image Formats and References\n\n"
    
    mu += "The following table describes these images: [1] their types and formats, [2] their quantities, [3] whether and how they are referenced from structured data such as TIU or Radiology notes, both directly and through groups, and [4] whether they go unreferenced (\"Orphan\") ...\n\n"
    
    mu += tbl.md() + "\n\n"
        
    mu += "<strong>IMAGE \"Images by Format and References\" - sb_imagesByFormatAndReferences - put in Office Excel and 2D Bar (Stacked) Chart ...\n\n"
    mu += "__VALUES__: {}\n\n".format(json.dumps(summs))
    mu += "![](/imgs/sb_imagesByFormatAndReferences.png)\n\n"
        
    """
    Orphans
    """
        
    mu += "Setting aside DICOM (.DCM), there are <span class='yellowIt'>{:,}</span> images, <span class='yellowIt'>{}</span> of which are _Orphans_.\n\n".format(allTotalLessDICOM, reportAbsAndPercent(allTotalOrphansLessDICOM, allTotalLessDICOM))
            
    mu += "Most _Orphans_ are accounted for by a property of most images called type index. In general, this property is ambiguous with most images ending up in catch-all categories such as _Medical Record_. However, many of its less used values do account for most of the Orphans.\n\n"

    # Reframe PDF under TI. Want to see TI that are all or mainly NOSET PDF.
    def groupTypeIndexes(subTypeInfos):
        tiUnsetGroupingByOT = defaultdict(lambda: defaultdict(int))
        tiGroupingByPDF = defaultdict(lambda: defaultdict(int))
        for subTypeInfo in subTypeInfos:
            if "type_index" not in subTypeInfo:
                continue
            otLabel = singleValue(subTypeInfo, "object_type").split(" [")[0]
            pdfLabel = singleValue(subTypeInfo, "pdf", "NOSET")
            for ti in subTypeInfo["type_index"]["byValueCount"]:
                tiGroupingByPDF[ti][pdfLabel] += subTypeInfo["type_index"]["byValueCount"][ti]
                if pdfLabel == "NOSET":
                    tiUnsetGroupingByOT[ti][otLabel] += subTypeInfo["type_index"]["byValueCount"][ti]
        return tiGroupingByPDF, tiUnsetGroupingByOT
    subTypeInfos = [subTypeInfo for ot in filerefSubTypesByOT for subTypeInfo in filerefSubTypesByOT[ot]] # if ot == "ADOBE"
    tiGroupingByPDF, tiUnsetGroupingByOT = groupTypeIndexes(subTypeInfos)
    majNoSetTIs = set()
    otherNoSetTIs = set()
    majNoSetOnlysTotal = 0
    otherNoSetOnlysTotals = 0
    for ti in tiGroupingByPDF:
        if "NOSET" not in tiGroupingByPDF[ti]:
            continue
        totalTI = sum(tiGroupingByPDF[ti][pdf] for pdf in tiGroupingByPDF[ti])
        noSetPercent = (float(tiGroupingByPDF[ti]["NOSET"])/float(totalTI)) * 100
        if noSetPercent >= 50.0: # > 50 percent then!
            majNoSetTIs.add(ti)
            majNoSetOnlysTotal += tiGroupingByPDF[ti]["NOSET"] # REM: don't want to cnt if any links
        else:
            otherNoSetTIs.add(ti)
            otherNoSetOnlysTotals += tiGroupingByPDF[ti]["NOSET"]
                
    mu += "Of <span class='yellowIt'>{:,}</span> active _type index_ values, <span class='yellowIt'>{:,}</span> account for <span class='yellowIt'>{:,}</span> Orphans ...\n\n".format(len(tiGroupingByPDF), len(majNoSetTIs), majNoSetOnlysTotal)
    
    tbl = MarkdownTable(["Type Index", "All/Majority Orphans", "Minority Referenced", "Orphan Type(s)"])
    for ti in sorted(list(majNoSetTIs), key=lambda x: tiGroupingByPDF[x]["NOSET"], reverse=True):
    
        setPDFMU = ", ".join(["{} ({})".format(pdfLabel, tiGroupingByPDF[ti][pdfLabel]) for pdfLabel in tiGroupingByPDF[ti] if pdfLabel != "NOSET"])
        otUSMU = ", ".join(["{} ({})".format(re.sub(r'\_', ' ', ot), tiUnsetGroupingByOT[ti][ot]) for ot in sorted(tiUnsetGroupingByOT[ti], key=lambda x: tiUnsetGroupingByOT[ti][x], reverse=True)]) if len(tiUnsetGroupingByOT[ti]) > 1 else tiUnsetGroupingByOT[ti].keys()[0]
        
        tbl.addRow([
            "__{}__".format(ti.split(" [")[0]), 
            tiGroupingByPDF[ti]["NOSET"], 
            setPDFMU, 
            otUSMU
        ])
        
    mu += tbl.md() + "\n\n"
    
    mu += "The balance of unreferenced Images, <span class='yellowIt'>{:,}</span>, belong to <span class='yellowIt'>{:,}</span> types that usually have references. For example, Progress Notes and Images are nearly always referenced. These Orphans are either mischaracterized or mistakenly left unreferenced.\n\n".format(otherNoSetOnlysTotals, len(otherNoSetTIs))
    
    tbl = MarkdownTable(["Type Index", "Minority Orphans", "Majority Referenced", "Orphan Type(s)"])
    for ti in sorted(list(otherNoSetTIs), key=lambda x: tiGroupingByPDF[x]["NOSET"], reverse=True):
    
        setPDFMU = ", ".join(["{} ({})".format(pdfLabel, tiGroupingByPDF[ti][pdfLabel]) for pdfLabel in tiGroupingByPDF[ti] if pdfLabel != "NOSET"])
        otUSMU = ", ".join(["{} ({})".format(re.sub(r'\_', ' ', ot), tiUnsetGroupingByOT[ti][ot]) for ot in sorted(tiUnsetGroupingByOT[ti], key=lambda x: tiUnsetGroupingByOT[ti][x], reverse=True)]) if len(tiUnsetGroupingByOT[ti]) > 1 else tiUnsetGroupingByOT[ti].keys()[0]
        cntAllOfTI = sum(tiGroupingByPDF[ti][pdfLabel] for pdfLabel in tiGroupingByPDF[ti])
        
        tbl.addRow([
            "__{}__".format(ti.split(" [")[0]), 
            reportAbsAndPercent(tiGroupingByPDF[ti]["NOSET"], cntAllOfTI),
            setPDFMU, 
            otUSMU
        ])
        
    mu += tbl.md() + "\n\n"
    
    """
    All possible referencing types 
    """ 
           
    tbl = MarkdownTable(["Id", "(Parent) Reference File", "Reference Subfile", "Abbrev", "Pkg", "Class"])
    noInThisSystemImages = 0
    resourceIter = filteredResultIterator(stationNo, "2005_03")
    for i, resource in enumerate(resourceIter, 1):
        # Chey 442 has corruption in its file ... 640 doesn't
        parentFileMU = "{} ({})".format(resource["file_pointer"]["label"], resource["file_pointer"]["id"].split("-")[1]) if resource["file_pointer"]["id"] != "1--1" else ""
        subFileMU = "{} ({})".format(resource["label"], resource["_id"].split("-")[1]) if resource["_id"].split("-")[1] != resource["file_pointer"]["id"].split("-")[1] else ""
        package_indexMU = "" if "package_index" not in resource else resource["package_index"].split(":")[1]
        class_indexMU = "" if "class_index" not in resource else resource["class_index"]["label"] # 2005_82
        root = resource["global_root_type"].split(":")[1]
        
        # Ignoring Root = [1] simple global (ex/ ^SRO), [2] global with fileid (ex/ ^XMB(3.9,) or [3] LAB with its multiples ie/ only ROOT 3 of interes
        # TODO: consider spec_subspec_index and proc_event_index (ie/ only other fields)
        
        inThisSystemsImages = resource["file_subfile_name"] in pdfLabelsSeen
        name = "__{}__".format(resource["file_subfile_name"]) if inThisSystemsImages else resource["file_subfile_name"]
        if inThisSystemsImages:
            noInThisSystemImages += 1
        
        tbl.addRow([
            name, 
            parentFileMU, 
            subFileMU, 
            resource["abbreviation"], 
            package_indexMU, 
            class_indexMU
        ])
                    
    mu += "VistA Imaging supports image references from <span class='yellowIt'>{:,}</span> types of record (\"parent data files\"). Many of these types - for example, Cardiac Catherization (691.1) - have been deprecated since VistA Imaging was launched and only a minority reference images today. This system employs only <span class='yellowIt'>{:,}</span> ...\n\n".format(i, noInThisSystemImages)
    
    mu += tbl.md() + "\n\n"
    
    return mu
    
"""
DICOM Image Report

FIRST OFF:
- redo last third for PUGET as must allow NONE for SOP too
- references: showing total indiv referenced vs unique TIU etc referencing ... gotta get at that (how?)
  - tie in types too
- nice if know roles involved + image first or doc first (preps for in general work)
- overlap cohorts
- ac devices
- location fixed or not

----- clean what's below here ----

Goal: ALL mand and key props exposed in a detailed dive
... REM: if fixed for some (RAD?) then want to SPLIT 2005 into logical sub types for DICOM/RAD vs others. ie/ what props relevant or not for different types of image work (TIU centric vs radiology note centric) ex/ pacs_procedure ONLY for rad_nuc ... (want to go to dataRF with 2005_X, 2005_Y)

Note: typer rerun 668 - added force on 74/radiology_report refs + group_parent refs + added status as prop <------- if works, must change args in base config + redo 663 too
... if same results => ie/ never see cross subTypeInfo linkage then go back to narrow
way (shouldn't see as 
NOTE: doing custom ... 8925 may be wrong as redoing within a date range but will be roughly right

TODO:
- 74's ... RangeCount + Real Count combo won't work (why group # != 74#?)
  <---- change code below to exception is not byValueCount for all!
- Group Parents as explicit too => see if really necessary AS should be unique for a set of images ie/ for spec_subspec_index 
  <------------ use explicit refs to a/c for why different? Do RAD NOTEs keep object ids?
<---------- ie want to know if 74 <----> Group is 1-1 and if not, why?
<---------- if you see the GROUP + ... CT Image ... seems like NOT to do with range combos (well if assume group and members share same SOP)
- see acquisition_site -- all ONE for biggies ie/ for RAD
- for RAD/NUC, is type_index always IMAGE?
- status for non RAD/NUC varies -- a/cs for TMP use of refs?
- note that patient # for DICOM is half overall patient # ie/ many have docs and no rad work?
- distinguish CLIN, DIAG and for Diag, Cardio, ECG etc. -----> comes from procedure?
- mand/opt props: MAND COUNT RN, NRN (see doing utils)
- XRAY GROUP other way for QA ie/ reference down. Should be broken by XRAY GROUP + sop class needed. See that the down reference works.
- DICOM GMRC TEMP LIST (http://localhost:9100/rambler#1-2006.5839) pathway ie/ independent
- ac device (ie/ round before going on to all the other images
- doc type (as do custom version typer) for the referenced dicoms!

TODO Utils:
- mand/opt props
- combine using subset of subtype props => if no byValueCount and > one use then exclude prop + combo those byValueCount + ie/ recombine properly
  <----- will greatly simplify code from broken down type
  
    TODO: going to own

https://wiki.cerner.com/display/VHAMigrationRequirements/Image+Migration+Details+-+Data+Volumes

FROM WIKI etc.            
    ... in the wiki, missing counts except for DICOM Studies
    ... have 34,820,989
    - got to break out cardiology as well as radiology
      - I think TIU DICOM?
    - they want to associate notes with images and images with notes so can relink after transfer (ADD BLURB to Image report on this)
    - images without TIU => will use patient ids and optionally encounter if available ... um!
    - REF: TIFF going to PDF ie/ TODO a format to linkage report
    - ECG is pending and not lumped with Radiology and Doc (but in Diag otherwise)
    - OR is cardiology rad reports and "Millenium will link to procedure/study"
      ... very focused on Rad Cardio Procedure/Study matching and criteria for it ... could do one ie/ do up front in pipeline
    - Use Case:
      Diagnostic Imaging
      - Radiology
      - Cardiovascular - Vascular, Echo, Cath Lab
      - Cardiology ECG
    ... TODO: see if images have any prop indicating Cardio/Rad?
      Clinical Imaging: no doc link, doc link
      Document Imaging
      (didn't capture non clinical images with no link!)
      <------ see if VA doc types differentiate at all?
      ... work to sprinkle in wiki
      
      Imaging Cohort is unclear (last three years seen or?)
"""
def webReportYR3DICOM(stationNo):

    print "FIRST CHECK: overlap bold alg wrong - had to manually fix in PDF + "\n" spaces and things run in as does TITLE[sop class oid]
    print x

    allThere, details = checkDataPresent(stationNo, [
    
        {"fileType": "2005", "check": "YR3"}, 
        {"fileType": "2006_532", "check": "ALL"} # SOP ids

    ])
    if not allThere:
        raise Exception("Some required data is missing - {}".format(details))

    allTypeData, subTypeDatas = splitTypeDatas(stationNo, "2005", reductionLabel="YR3", expectSubTypeProperty="object_type-capture_application-dicom_sop_class-spec_subspec_index-status")
    
    overallFirstCreateDate = allTypeData["_firstCreateDate"].split("T")[0] 
    overallLastCreateDate = allTypeData["_lastCreateDate"].split("T")[0]
    
    dicomSOPClassNameByUID = dict((resource["sop_class_uid"], resource["sop_class_name"]) for resource in filteredResultIterator(stationNo, "2006_532"))
    
    subTypeInfosByOT = defaultdict(list)
    for subTypeInfo in subTypeDatas:
        ot = singleValue(subTypeInfo, "object_type").split(" [")[0]
        subTypeInfosByOT[ot].append(subTypeInfo)
        
    totalDICOM = 0
    sopClassCategoryCount = Counter()
    allSOPClassed = True
    patients = set()
        
    totalDICOMRN = 0 
    countBySpecIndexBySOPClassRN = defaultdict(lambda: Counter())
    patientRN = set()
    propCountsRN = Counter()
    radRepRefsBySOPClass = defaultdict(lambda: set())
    countGroupRefsBySOPClassRN = Counter() 
    countByPacsProcBySOPClass = defaultdict(lambda: Counter()) # only RAD/NUC

    totalDICOMNRN = 0 
    patientNRN = set()
    propCountsNRN = Counter()
    countBySpecIndexBySOPClassNRN = defaultdict(lambda: Counter())
    countGroupRefsBySOPClassNRN = Counter()
    countByPDFBySOPClassNRN = defaultdict(lambda: Counter()) # only non RAD/NUC as ambig

    # All 1.2.840.10008.5.1.4.1.1.77.1.4 ie/ non RAD/NUC, not DICOM Gateway but Capture Workstation)
    totalDICOMNRNCA = 0   
    patientNRNCA = set() 
    propCountsNRNCA = Counter()
    countBySpecIndexBySOPClassNRNCA = defaultdict(lambda: Counter())
    countGroupRefsBySOPClassNRNCA = Counter()
    countByPDFBySOPClassNRNCA = defaultdict(lambda: Counter())
    countByStatusBySOPClassNRNCA = defaultdict(lambda: Counter())
    
    for subTypeInfo in subTypeInfosByOT["DICOM IMAGE"]:

        # REM: split by it so all and 1!
        if "dicom_sop_class" in subTypeInfo:
            sopClassRef = subTypeInfo["dicom_sop_class"]["byValueCount"].keys()[0] # only one
            sopClassUID = sopClassRef.split(" [")[0]
        else:
            sopClassUID = "NOTSET"
            allSOPClassed = False

        if "spec_subspec_index" not in subTypeInfo:
            specSubSpecIndexLabel = "NOTSET"
        else:
            if subTypeInfo["spec_subspec_index"]["count"] != subTypeInfo["_total"]:
                raise Exception("If spec_subspec_index for SOP then except for all of SOP")
            specSubSpecIndexLabel = subTypeInfo["spec_subspec_index"]["byValueCount"].keys()[0].split(" [")[0]
                            
        """
        RADIOLOGY and NUCLEAR MEDICINE 
          * always and only ones ref'ed by radiology_report
          * has pacs_procedure (and others?) too?
          * status is ALWAYs Viewable
        
        Enforce:
            > if spec_subspec_index is RADIOLOGY or NUCLEAR MEDICINE then expect radiology_report for all (see range), parent_data_file of RADIOLOGY [2005_03...] ( radiology_report should equal 74 + parent_global_root_0) AND capture_application is always DICOM_Gateway and there is a pacs_procedure
        
        Note: may have > #  of group_parent's (2005's) than radiology_reports => > 1
        group possible. But should never have more groups than radiology reports 
        """
        if specSubSpecIndexLabel in ["RADIOLOGY", "NUCLEAR MEDICINE"]:
        
            if not (
            
                "parent_data_file" in subTypeInfo and
                subTypeInfo["parent_data_file"]["count"] == subTypeInfo["_total"] and
                len(subTypeInfo["parent_data_file"]["byValueCount"]) == 1 and
                re.match(r'RADIOLOGY', subTypeInfo["parent_data_file"]["byValueCount"].keys()[0]) and
                
                "radiology_report" in subTypeInfo and
                subTypeInfo["radiology_report"]["count"] == subTypeInfo["_total"] and
                
                "capture_application" in subTypeInfo and             
                len(subTypeInfo["capture_application"]["byValueCount"]) == 1 and 
                subTypeInfo["capture_application"]["byValueCount"].keys()[0] == "D:DICOM Gateway" and
                
                "pacs_procedure" in subTypeInfo and
                subTypeInfo["pacs_procedure"]["count"] == subTypeInfo["_total"] and
                
                "group_parent" in subTypeInfo and
                subTypeInfo["group_parent"]["count"] == subTypeInfo["_total"] and
                
                "status" in subTypeInfo and
                set(subTypeInfo["status"]["byValueCount"].keys()) == set(["1:Viewable"])
                
                # for RAD/NUC, is type_index always IMAGE?
                # is class_index always CLIN?
                
                # Agreement of D0 spread (if there) and group_parent and radiology_report ie/ #'s if byValueCounted should agree
                
            ):
                raise Exception("If spec_subspec_index is RADIOLOGY or NUCLEAR MEDICINE then expect a radiology_report for all members, status Viewable (as DICOM Gateway), pacs procedure and other things")
                
            sopClassCategoryCount[sopClassUID] += 1 # for overlaps
                
            for prop in subTypeInfo:
                if re.match(r'_', prop):
                    continue
                propCountsRN[prop] += subTypeInfo[prop]["count"]
            
            # total DICOM Rad Nuc (vs others)
            totalDICOMRN += subTypeInfo["_total"]
            totalDICOM += subTypeInfo["_total"]
            
            for patientRef in subTypeInfo["patient"]["byValueCount"]:
                patientRN.add(patientRef)
                patients.add(patientRef)
            
            countBySpecIndexBySOPClassRN[sopClassUID][specSubSpecIndexLabel] += subTypeInfo["_total"]
                        
            radRepRefsBySOPClass[sopClassUID] |= set(subTypeInfo["radiology_report"]["byValueCount"])
            
            bvc = subTypeInfo["pacs_procedure"]["byValueCount"]
            for pacsProc in bvc:
                countByPacsProcBySOPClass[sopClassUID][pacsProc.split(" [")[0]] += bvc[pacsProc]
                
            # Note: TODO - change to being set like rad refs as MUST be explicit
            countGroupRefsBySOPClassRN[sopClassUID] += subTypeInfo["group_parent"]["rangeCount"] if "rangeCount" in subTypeInfo["group_parent"] else len(subTypeInfo["group_parent"]["byValueCount"])
                           
        else:
        
            # All not RAD/NUC: no rad rep or pacs
            if (
                "radiology_report" in subTypeInfo or
                "pacs_procedure" in subTypeInfo
            ):
                raise Exception("Don't expect a radiology_report reference or pacs_procedure for any images not in RADIOLOGY/NUCLEAR MEDICINE spec_subspec_index")  
        
            """
            Second: not RAD/NUC but from DICOM Gateway
            """
            if re.search(r'DICOM_Gateway', subTypeInfo["_subTypeId"]): # NRN (all G/W) 
            
                if not (
                    "status" in subTypeInfo and
                    set(subTypeInfo["status"]["byValueCount"].keys()) == set(["1:Viewable"])
                ):
                    raise Exception("Expected 2: non RAD/NUC but DICOM Gateway to always have status Viewable ie G/W import is fixed to Viewable")
            
                sopClassCategoryCount[sopClassUID] += 1 # for overlaps
                
                for prop in subTypeInfo:
                    if re.match(r'_', prop):
                        continue
                    propCountsNRN[prop] += subTypeInfo[prop]["count"]
                
                totalDICOMNRN += subTypeInfo["_total"]
                totalDICOM += subTypeInfo["_total"]
            
                for patientRef in subTypeInfo["patient"]["byValueCount"]:
                    patientNRN.add(patientRef)
                    patients.add(patientRef)
            
                countBySpecIndexBySOPClassNRN[sopClassUID][specSubSpecIndexLabel] += subTypeInfo["_total"]

                # Note: assuming separate groups if separate SPEC INDEX 
                countGroupRefsBySOPClassNRN[sopClassUID] += subTypeInfo["group_parent"]["rangeCount"] if "rangeCount" in subTypeInfo["group_parent"] else len(subTypeInfo["group_parent"]["byValueCount"])
            
                if "parent_data_file" in subTypeInfo:
                    bvc = subTypeInfo["parent_data_file"]["byValueCount"]
                    for pdf in bvc:
                        countByPDFBySOPClassNRN[sopClassUID][pdf.split(" [")[0]] += bvc[pdf]
                else:
                    countByPDFBySOPClassNRN[sopClassUID]["NOTSET"] += subTypeInfo["_total"]
                
            # Third: not RAD/NUC and (NOT from DICOM Gateway) Capture Application
            # - Mix of status: Viewable but others too
            # - Only SOP UID (which is set) fits here
            #      VL Photographic Image Storage [1.2.840.10008.5.1.4.1.1.77.1.4]
            else: # NRNCA
            
                if not (
                    sopClassUID in ["1.2.840.10008.5.1.4.1.1.77.1.4", "NOTSET"] and
                    
                    singleValue(subTypeInfo, "capture_application", "NOTSET") == "C:Capture Workstation"
                ):
                    raise Exception("Expect all NRNCA to be UID 1.2.840.10008.5.1.4.1.1.77.1.4 or NONE and as not DICOM Gateway to be CAPTURE APPLICATION")
            
                sopClassCategoryCount[sopClassUID] += 1 # for overlaps
                
                for prop in subTypeInfo:
                    if re.match(r'_', prop):
                        continue
                    propCountsNRNCA[prop] += subTypeInfo[prop]["count"]
                
                totalDICOMNRNCA += subTypeInfo["_total"]
                totalDICOM += subTypeInfo["_total"]
                
                for patientRef in subTypeInfo["patient"]["byValueCount"]:
                    patientNRNCA.add(patientRef)
                    patients.add(patientRef)
            
                countBySpecIndexBySOPClassNRNCA[sopClassUID][specSubSpecIndexLabel] += subTypeInfo["_total"]

                # Note: assuming separate groups if separate SPEC INDEX 
                # Saw missing in PUG - seemed to be a test image(s) 
                if "group_parent" in subTypeInfo: 
                    countGroupRefsBySOPClassNRNCA[sopClassUID] += subTypeInfo["group_parent"]["rangeCount"] if "rangeCount" in subTypeInfo["group_parent"] else len(subTypeInfo["group_parent"]["byValueCount"])
            
                if "parent_data_file" in subTypeInfo:
                    bvc = subTypeInfo["parent_data_file"]["byValueCount"]
                    for pdf in bvc:
                        countByPDFBySOPClassNRNCA[sopClassUID][pdf.split(" [")[0]] += bvc[pdf]
                else:
                    countByPDFBySOPClassNRNCA[sopClassUID]["NOTSET"] += subTypeInfo["_total"]
                    
                status = singleValue(subTypeInfo, "status", "NOTSET")
                countByStatusBySOPClassNRNCA[sopClassUID][status] += subTypeInfo["status"]["byValueCount"][status]
                
    userSiteDir = SITE_DIR_TEMPL.format(stationNo)
    if not os.path.isdir(userSiteDir):
        raise Exception("Expect User Site to already exist with its basic contents")
        
    title = "Imaging [DICOM]".format(stationNo)
    mu = TOP_MD_TEMPL.format(title, "DICOM Images")
                                    
    sopClassCoverageQualifier = ""
    if not allSOPClassed:
        sopClassCoverageQualifier = " nearly "    
    mu += """Of the <span class='yellowIt'>{:,}</span> images added to this VistA between _{}_ and _{}_, a three year span, the <span class='yellowIt'>{}</span> DICOM Images represent the vast majority. They are for <span class='yellowIt'>{:,}</span> patients and {}all are tagged with a \"SOP Class\". Three categories of DICOM image are detailed below and note that should a SOP Class belong to more than one category, its name is rendered in bold. 
 
""".format(
        allTypeData["_total"],
        overallFirstCreateDate,
        overallLastCreateDate,
        reportAbsAndPercent(totalDICOM, allTypeData["_total"]),
        len(patients),
        sopClassCoverageQualifier
    )
    
    # Bolds if class appears in more than one of the three categories
    def muSOPClass(sopClassUID):
        if sopClassUID != "NOTSET":
            labelMU = "__{}__".format(dicomSOPClassNameByUID[sopClassUID]) if sopClassCategoryCount[sopClassUID] > 1 else dicomSOPClassNameByUID[sopClassUID]
            labelMU = "{} [{}]".format(labelMU, sopClassUID)
        else:
            labelMU = "NOTSET"
        return labelMU
        
    def muCounter(counter):
        if len(counter) > 1:
            cmu = ", ".join(["{} [{:,}]".format(k.split(":")[1] if re.search(r':', k) else k, counter[k]) for k in sorted(counter, key=lambda x: counter[x], reverse=True)])
        else:
            k = counter.keys()[0]
            cmu = k if not re.search(r':', k) else k.split(":")[1]      
        return cmu
        
    """
    First: RAD/NUC (and DICOM Gateway)
    """
    mu += """The majority of DICOM Images, <span class='yellowIt'>{}</span>, are classed as _RADIOLOGY_ or _NUCLEAR MEDICINE_ and cover <span class='yellowIt'>{}</span> patients. All are referenced from Radiology Reports and imported using _DICOM Gateway_. A \"pacs procedure\" is specified for all these images. Note that apart from references from Radiology Reports, the System maintains these images in groups (\"XRAY GROUP\") ..."
    
""".format(
        reportAbsAndPercent(totalDICOMRN, totalDICOM),
        reportAbsAndPercent(len(patientRN), len(patients))
    )
    tbl = MarkdownTable(["SOP Class", "Count", "Class(es)", "Radiology Report Referrers", "Groups", "PACS Procedure Types"])
    for sopClassUID in sorted(countBySpecIndexBySOPClassRN, key=lambda x: sum(countBySpecIndexBySOPClassRN[x][y] for y in countBySpecIndexBySOPClassRN[x]), reverse=True):
        thisTotal = sum(countBySpecIndexBySOPClassRN[sopClassUID][x] for x in countBySpecIndexBySOPClassRN[sopClassUID])
        tbl.addRow([
            muSOPClass(sopClassUID),
            reportAbsAndPercent(thisTotal, totalDICOMRN),
            muCounter(countBySpecIndexBySOPClassRN[sopClassUID]),
            len(radRepRefsBySOPClass[sopClassUID]),
            countGroupRefsBySOPClassRN[sopClassUID],
            len(countByPacsProcBySOPClass[sopClassUID])
        ])
    mu += tbl.md() + "\n\n"
    
    """
    Second: not RAD/NUC but from DICOM Gateway
    """
    mu += """Most other DICOM Images, <span class='yellowIt'>{}</span>, are also entered through the DICOM Gateway and cover <span class='yellowIt'>{}</span> patients. They either go unreferenced, are referenced by a temporary holding file (\"DICOM GMRC TEMP LIST\") or from TIU Notes. They cover Ophthalmology, Dentistry and Vascular measurement. Some of these SOP Classes (in bold) are also used for the Radiology images detailed above ...
    
""".format(
        reportAbsAndPercent(totalDICOMNRN, totalDICOM),
        reportAbsAndPercent(len(patientNRN), len(patients))
    )
    tbl = MarkdownTable(["SOP Class", "Count", "Class(es)", "Referrers", "Groups"])
    for sopClassUID in sorted(countBySpecIndexBySOPClassNRN, key=lambda x: sum(countBySpecIndexBySOPClassNRN[x][y] for y in countBySpecIndexBySOPClassNRN[x]), reverse=True):
    
        thisTotal = sum(countBySpecIndexBySOPClassNRN[sopClassUID][x] for x in countBySpecIndexBySOPClassNRN[sopClassUID])
                
        tbl.addRow([
            muSOPClass(sopClassUID),
            reportAbsAndPercent(thisTotal, totalDICOMNRN),
            muCounter(countBySpecIndexBySOPClassNRN[sopClassUID]),
            muCounter(countByPDFBySOPClassNRN[sopClassUID]),
            countGroupRefsBySOPClassNRN[sopClassUID]
        ])
    mu += tbl.md() + "\n\n"
    
    """
    Third: Not RAD/NUC and not from DICOM Gateway
    
    Either SOP 1.2.840.10008.5.1.4.1.1.77.1.4 or NO SOP 
    """
    if set(countBySpecIndexBySOPClassNRNCA.keys()) == set(["1.2.840.10008.5.1.4.1.1.77.1.4"]):
        sopClassMsg = "They all share the same SOP Class, _VL Photographic Image Storage [1.2.840.10008.5.1.4.1.1.77.1.4]_"
    else:
        sopClassMsg = "Where SOP Class is set, it is always, _VL Photographic Image Storage [1.2.840.10008.5.1.4.1.1.77.1.4]_"
    mu += """The remaining DICOM Images, <span class='yellowIt'>{}</span>, are imported using \"Capture Application\" and cover <span class='yellowIt'>{}</span> patients. {}. References are like those of their non RADIOLOGY peers but these images have a tracked status value too. That status captures whether they are viewable of not ..."
    
""".format(
        reportAbsAndPercent(totalDICOMNRNCA, totalDICOM),
        reportAbsAndPercent(len(patientNRNCA), len(patients)),
        sopClassMsg
    )
    
    tbl = MarkdownTable(["SOP Class", "Count", "Class(es)", "Referrers", "Groups", "Status"], includeNo=False)
    for sopClassUID in sorted(countBySpecIndexBySOPClassNRNCA, key=lambda x: sum(countBySpecIndexBySOPClassNRNCA[x][y] for y in countBySpecIndexBySOPClassNRNCA[x]), reverse=True):
    
        thisTotal = sum(countBySpecIndexBySOPClassNRNCA[sopClassUID][x] for x in countBySpecIndexBySOPClassNRNCA[sopClassUID])
                
        tbl.addRow([
            muSOPClass(sopClassUID),
            reportAbsAndPercent(thisTotal, totalDICOMNRNCA),
            muCounter(countBySpecIndexBySOPClassNRNCA[sopClassUID]),
            muCounter(countByPDFBySOPClassNRNCA[sopClassUID]),
            countGroupRefsBySOPClassNRNCA[sopClassUID],
            muCounter(countByStatusBySOPClassNRNCA[sopClassUID])
        ])
    mu += tbl.md() + "\n\n"
    
    """
    TODO: overlap section - give overlap with other known systems but qualify when those were cut. Give exclusive to this and its overlaps and total overall.
    """
    
    """
    TODO: AC DEV's 
    
    and note dominantly [3] is Dermatology ...
    
    VL Photographic Image Storage [1.2.840.10008.5.1.4.1.1.77.1.4]
    """
    
    """
    TODO: acquisition_site
    """
     
    # Output   
    userSiteDir = SITE_DIR_TEMPL.format(stationNo)
    open(userSiteDir + "imagingDICOM.md", "w").write(mu)
    
"""
Beyond DICOM
"""
def webReportYR3Other(stationNo): 
    allTypeData, subTypeDatas = splitTypeDatas(stationNo, "2005", reductionLabel="YR3", expectSubTypeProperty="object_type-capture_application-dicom_sop_class-spec_subspec_index")
    
# ################################# DRIVER #######################
               
def main():

    assert(sys.version_info >= (2,7))

    if len(sys.argv) < 2:
        print "need to specify station # ex/ 442 - exiting"
        return
        
    stationNo = sys.argv[1]
    
    userSiteDir = SITE_DIR_TEMPL.format(stationNo)
    if not os.path.isdir(userSiteDir):
        raise Exception("Expect User Site to already exist with its basic contents")

    webReportImaging(stationNo)
                 
if __name__ == "__main__":
    main()



from setuptools import setup

setup(name='fmqlreports',
      description = 'FMQL Reports',
      long_description = """A Python framework and set of executables for generating specific reports on VistA data cached using FMQL.""",
      version='0.39',
      install_requires = [
          'fmqlutils>=2.92'
          # 'numpy>=1.16.3'
      ],
      classifiers = ["Development Status :: 4 - Beta", "Programming Language :: Python :: 2.7"],
      url='http://github.com/Caregraf/FMQL/fmqlreports',
      license='Apache License, Version 2.0',
      keywords='VistA,FileMan,CHCS,FMQL,Caregraf',
      # install_requires=["pytz"], - problem OSX 
      package_dir = {'fmqlreports': ''},
      packages = ['fmqlreports', 'fmqlreports.basics', 'fmqlreports.patient', 'fmqlreports.user', 'fmqlreports.schema', 'fmqlreports.imaging'],
      entry_points = {
      }
)



#!/usr/bin/env python
# -*- coding: utf8 -*-

import sys
import os
import re
import json
from collections import defaultdict, Counter
from datetime import datetime

from fmqlutils.reporter.reportUtils import MarkdownTable, reportPercent, reportAbsAndPercent

from fmqlutils.typer.reduceTypeUtils import splitTypeDatas, checkDataPresent, singleValue

from fmqlreports.webReportUtils import SITE_DIR_TEMPL

"""
Basic Patient

   - # total
   - # alive, dead
   - longevity median
   - median age of those alive
   - veteran y or n for alive
   - period of service for all
   (ie/ quick go)
"""
def webReportPatient(stationNo):

    allThere, details = checkDataPresent(stationNo, [
        {"fileType": "2", "check": "TYPE"}
    ])
    if not allThere:
        raise Exception("Some required data is missing - {}".format(details))

    mu = """---
layout: default
title: {} Patient
---

## Patient Basics

""".format(stationNo) 

    # Practice of Care    
    type2, subs = splitTypeDatas(stationNo, "2", expectSubTypeProperty="date_of_death")
    
    print "Total known {}".format(type2["_total"])
    print "Alive {}".format(type2["_total"] - type2["date_of_death"]["count"])
    longevity = Counter()
    for sub in subs:
        if "date_of_death" not in sub:
            aliveSub = sub
            continue
        deathDate = int(singleValue(sub, "date_of_death"))
        for dob in sub["date_of_birth"]["byValueCount"]:
            ageAtDeath = deathDate - int(dob)
            longevity[ageAtDeath] += 1
    print longevity
        
    print aliveSub["period_of_service"]
    print 
    return
    
    userSiteDir = SITE_DIR_TEMPL.format(stationNo)
    open(userSiteDir + "patient.md", "w").write(mu)

"""
Criteria: if has entry for ANY of vitals, ... in last three years.

Required: SO 3 Years for key clinical data

TODO: move to webReportPatientUtils.py and may write to TMP
"""    
def lastThreeYearsPatients(stationNo):
    pass
        
# ################################# DRIVER #######################
               
def main():

    assert(sys.version_info >= (2,7))
    
    if len(sys.argv) < 2:
        print "need to specify station # ex/ 442 - exiting"
        return
        
    stationNo = sys.argv[1]
    
    userSiteDir = SITE_DIR_TEMPL.format(stationNo)
    if not os.path.isdir(userSiteDir):
        raise Exception("Expect User Site to already exist with its basic contents")
    
    webReportPatient(stationNo)
    
if __name__ == "__main__":
    main()
